/*
 * sbminst.c
 *
 * the installation program for Smart BootManager
 * 
 * this program only supports Linux and DOS.
 *
 * Copyright (C) 2000, Suzhe, Lonius  See file COPYING for details.
*/



#if ( !defined( __linux__ ) && !defined( __DJGPP__ ) ) || !defined( __GNUC__ )
#error "This program only can be compiled by GNU C under Linux or DJGPP"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <errno.h>
#include <ctype.h>

#ifdef __linux__

#include <linux/fs.h>
#include <linux/hdreg.h>
#include <linux/fd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>

#else

#include <bios.h>
#include <dos.h>

#endif

#include "loader.h"
#include "main.h"
#include "theme-us.h"
#include "theme-zh.h"
#include "sbminst.h"

#ifdef __linux__
char ldr_dev[20], knl_dev[20];
struct stat st;
#else
int ldev_id, kdev_id;
byte dev_sectors, dev_heads;
#endif

int knl_tsize;
FILE *bk_fp, *ui_fp;
byte buf_ldr[SECTOR_SIZE + 1], buf_knl[MAX_SBM_SIZE];
byte buf_theme[MAX_THEME_SIZE];
byte *theme_code;

int always_yes;

struct sbml_header *p_loader;
struct sbmk_header *p_knl;
struct sbmt_header *p_theme;

void
init_settings ()
{
  knl_tsize = 0;		/* kernel part total size */
  bk_fp = NULL;
  ui_fp = NULL;
  theme_code = theme_us_code;
  always_yes = 0;

#ifdef __linux__
  strcpy(ldr_dev, "/dev/fd0");
  strcpy(knl_dev, "/dev/fd0");
#else
  ldev_id = kdev_id = 0;
#endif
  
}

void
die (char *fmt, ...)
{
  va_list ap;
  va_start (ap, fmt);
  vfprintf (stderr, fmt, ap);
  va_end (ap);
  fprintf (stderr, "\n");
  exit (-1);
}

static void
usage (char *name)
{
  char *here;

  here = strrchr (name, '/');
  if (here)
    name = here + 1;
  fprintf (stderr, "usage:\n %s %s", name, usage_msg);
  exit (0);
}

byte
calc_checksum (byte * knl, int size)
{
  byte checksum;
  int i;
  for (i = 0, checksum = 0; i < size; i++)
    checksum += (*knl++);
  checksum = -((char) checksum);
  return checksum;
}

void
grasp_params (int num, char *index[])
{
  int i, theme_size;
  FILE *theme_fp;
  char *name;

  i = num--;
  name = *index;
  if (!num || index[1][0] != '-')
    usage (name);

  while (num && (*(++index))[0] == '-')
    {
      if ((*index)[2])
	usage (name);
      num--;
      switch ((*index)[1])
	{
	case 't':
	  if (!num)
	    {
	      printf ("\nincomplete option %s\n", (*index));
	      usage (name);
	    }
	  num--, index++;
	  if (strcmp ((*index), "zh") == 0)
	    theme_code = theme_zh_code;
	  else if (strcmp ((*index), "us") == 0)
	    theme_code = theme_us_code;
	  else
	    {
	      if ((theme_fp = fopen ((*index), "rb")) == NULL)
		die ("fopen %s: %s", (*index), strerror (errno));
	      fseek (theme_fp, 0, SEEK_END);
	      theme_size = ftell (theme_fp);
	      rewind (theme_fp);

	      if (fread (buf_theme, (theme_size > 16384) ? 16384 : theme_size,
			 1, theme_fp) < 1)
		die ("fread %s: %s", (*index), strerror (errno));
	      theme_code = buf_theme;
	      fclose (theme_fp);
	    }
	  break;
	case 'd':
	  if (!num)
	    {
	      printf ("\nincomplete option %s\n", (*index));
	      usage (name);
	    }
	  num--, index++;

#ifdef __linux__
	  strncpy (ldr_dev, (*index), 18);
	  ldr_dev[19] = '\0';
	  strcpy (knl_dev, ldr_dev);
#else
          ldev_id = atoi (*(index));
          kdev_id = ldev_id;
#endif

	  break;

	  /*case  'k':
	     if (!num) usage(name);
	     num --, index ++;
	     printf("Are you sure to install the kernel into %s? [y/n] ", (*index));
	     fgets(input, 2, stdin);
	     printf("%s\n", input);
	     if (input[0] != 'y')
	     exit(-1);
	     strncpy(knl_dev, (*index), 9);
	     knl_dev[9] = '\0';
	     break;
	   */

	case 'b':
	  if (!num)
	    {
	      printf ("\nincomplete option %s\n", (*index));
	      usage (name);
	    }
	  num--, index++;
	  if ((bk_fp = fopen ((*index), "wb")) == NULL)
	    die ("fopen %s : %s", (*index), strerror (errno));
	  break;
	case 'u':
	  if (!num)
	    {
	      printf ("\nincomplete option %s\n", (*index));
	      usage (name);
	    }
	  num--, index++;
	  if (i > 3)
	    die ("uninstall option can only be used solely.");
	  if ((ui_fp = fopen ((*index), "rb")) == NULL)
	    die ("fopen %s : %s", (*index), strerror (errno));
	  break;
	case 'y':
	  always_yes = 1;
	  break;
	default:
	  usage (name);
	}
    }

#ifdef DEBUG
  printf ("end of grasping the params.\n");
#endif

}

#ifdef __linux__

int
open_blk_dev (char *dev)
{
  int devd;
  if ((devd = open (dev, O_RDWR)) < 0)
    die ("open %s: %s", dev, strerror (errno));
  if (fstat (devd, &st) < 0)
    die ("stat %s: %s", dev, strerror (errno));
  if (!S_ISBLK (st.st_mode))
    die ("%s is not a block device!\n", dev);

#ifdef DEBUG
  printf ("open %s for R/W, press any key to continue...", dev);
  getchar ();
#endif
  return devd;
}

void
close_blk_dev (int *dev_d)
{
  sync ();
  if (close (*dev_d) < 0)
    die ("close %s : %s", knl_dev, strerror (errno));
  *dev_d = 0;
}

#else

void
check_dev (int drv_id)
{
  byte tmp[10];
  if (biosdisk (8, drv_id, 0, 0, 1, 1, tmp))
    die ("Disk check error!");

  dev_sectors = tmp[0] & 0x3f;
  dev_heads = tmp[3];

  if (!dev_sectors || !dev_heads)
    die ("Disk check error!");
}

void
disk_io (int func, byte dev_id, int head, int cyl, int sect, char *buf,
         int sum)
{
  int i;
  for (i = 0; i < sum; i++)
    {
      if (biosdisk (func, dev_id, head, cyl, sect, 1, buf + i * 512))
        die ("Disk read/write error!");
      sect++;
      if (sect > dev_sectors)
        {
          sect = 1;
          if (head >= dev_heads)
            {
              head = 0;
              cyl++;
            }
          else
            head++;
        }
    }
}

#endif

void
proc_knl ()
{

  int devd, sect, head, cyl;

#ifdef __linux__
  devd = open_blk_dev (knl_dev);
#else
  check_dev (kdev_id);
#endif

  memcpy (buf_knl, main_code, p_knl->kernel_size);
  memcpy (buf_knl + p_knl->kernel_size, theme_code, p_theme->size);

  p_knl = (struct sbmk_header *) buf_knl;

  p_knl->total_sectors = (knl_tsize + 511) / SECTOR_SIZE;
  p_knl->kernel_copies = 1;
  p_knl->kernel_addr1 = KNL_LBA;

#ifdef __linux__
  if (MAJOR (st.st_rdev) == MAJOR_FD)
    p_knl->kernel_drvid1 = MINOR (st.st_rdev) & 3;
  else if (MAJOR (st.st_rdev) == MAJOR_HD)
    p_knl->kernel_drvid1 = 0x80 + (MINOR (st.st_rdev) >> 6);
  else
    die ("device %s not support in this version.", knl_dev);
#else
  p_knl->kernel_drvid1 = kdev_id;
#endif

#ifdef DEBUG
  printf ("kernel drive_id = 0x%x\n", p_knl->kernel_drvid1);
#endif

  p_knl->checksum = 0;
  p_knl->checksum = calc_checksum (buf_knl, p_knl->checksum_size);


  /* always begin at safe place */
#ifdef __linux__
  lseek (devd, SECTOR_SIZE, SEEK_SET);
  if (write (devd, buf_knl, knl_tsize) != knl_tsize)
    die ("write %s: %s", knl_dev, strerror (errno));

  close_blk_dev (&devd);
#else
  disk_io (3, kdev_id, KNL_HEAD, KNL_CYL, KNL_SECT, buf_knl,
           p_knl->total_sectors);
#endif

#ifdef DEBUG
  printf ("Kernel part installed successfully.\n");
#endif
}

void
proc_ldr ()
{
  char *s;
  int devd;

#ifdef __linux__
  devd = open_blk_dev (ldr_dev);
#else
  check_dev (ldev_id);
  disk_io(2, ldev_id, LDR_HEAD, LDR_CYL, LDR_SECT, buf_ldr, 1);
#endif

  memcpy (buf_ldr, loader_code, LDR_SIZE);

  p_loader->kernel_drvid1 = p_knl->kernel_drvid1;
  p_loader->kernel_copies = 1;
  p_loader->kernel_addr1 = KNL_LBA;
  p_loader->kernel_sectors = p_knl->total_sectors;

#ifdef DEBUG
  printf ("loader drive_id = 0x%x\n", p_loader->kernel_drvid1);
#endif

#ifdef __linux__
  lseek (devd, 0, SEEK_SET);
  if (write (devd, buf_ldr, LDR_SIZE) != LDR_SIZE)
    die ("write %s: %s", ldr_dev, strerror (errno));

  close_blk_dev (&devd);
#else
  disk_io (3, ldev_id, LDR_HEAD, LDR_CYL, LDR_SECT, buf_ldr, 1);
#endif

#ifdef DEBUG
  printf ("Loader part installed successfully.\n");
#endif

}

void
backup ()
{
  int devd, ksectors;

  if (bk_fp == NULL)
    return;

  printf("Backuping the data...\n");

#ifdef __linux__

  devd = open_blk_dev (ldr_dev);
  lseek (devd, 0, SEEK_SET);

  if (read (devd, buf_ldr, SECTOR_SIZE) != SECTOR_SIZE)
    die ("read %s: %s", ldr_dev, strerror (errno));

  if (fwrite (BACKUP_MAGIC, strlen(BACKUP_MAGIC), 1, bk_fp) < 1 ||
      fwrite (ldr_dev, 20, 1, bk_fp) < 1 ||
      fwrite (buf_ldr, SECTOR_SIZE, 1, bk_fp) < 1 )
    die ("fwrite backup file error");

  close_blk_dev (&devd);

  devd = open_blk_dev (knl_dev);
  lseek (devd, SECTOR_SIZE, SEEK_SET);
  if (read (devd, buf_knl, knl_tsize) != knl_tsize)
    die ("read %s: %s", knl_dev, strerror (errno));

  if (fwrite (knl_dev, 20, 1, bk_fp) < 1 ||
      fwrite (&knl_tsize, sizeof (knl_tsize), 1, bk_fp) < 1 ||
      fwrite (buf_knl, knl_tsize, 1, bk_fp) < 1 )
    die ("fwrite backup file error");

  close_blk_dev (&devd);

#else

  check_dev (ldev_id);

  disk_io (2, ldev_id, LDR_HEAD, LDR_CYL, LDR_SECT, buf_ldr, 1);

  if (fwrite (BACKUP_MAGIC, strlen (BACKUP_MAGIC), 1, bk_fp) < 1 ||
      fwrite (&ldev_id, sizeof (ldev_id), 1, bk_fp) < 1 ||
      fwrite (buf_ldr, SECTOR_SIZE, 1, bk_fp) < 1)
    die ("write backup file error");

  ksectors = (knl_tsize + SECTOR_SIZE - 1) / SECTOR_SIZE;
  check_dev (kdev_id);
  disk_io (2, kdev_id, KNL_HEAD, KNL_CYL, KNL_SECT, buf_knl, ksectors);

  if (fwrite (&kdev_id, sizeof (kdev_id), 1, bk_fp) < 1 ||
      fwrite (&ksectors, sizeof (ksectors), 1, bk_fp) < 1 ||
      fwrite (buf_knl, ksectors * SECTOR_SIZE, 1, bk_fp) < 1)
    die ("write backup file error");

#endif

  fclose (bk_fp);
}

void
unistall ()
{
  int ks;
  int devd;
  char buf[20];

  if (!always_yes)
    {
      printf ("Are you really want to uninstall Smart BootManager [y/n]?");
      fgets (buf, 19, stdin);

      if (buf[0] != 'y')
	die ("Abort uninstallation.");
    }

  if (fread (buf, strlen(BACKUP_MAGIC), 1, ui_fp) < 1)
    die ("fread backup file error");
  if (strcmp (buf, BACKUP_MAGIC) != 0)
    die ("Invalid backup file");

#ifdef __linux__

  if (fread (ldr_dev, 20, 1, ui_fp) < 1 ||
      fread (buf_ldr, SECTOR_SIZE, 1, ui_fp) < 1 ||
      fread (knl_dev, 20, 1, ui_fp) < 1 ||
      fread (&ks, sizeof (ks), 1, ui_fp) < 1 ||
      fread (buf_knl, ks, 1, ui_fp) < 1 )
    die ("fread backup file error");

  devd = open_blk_dev (ldr_dev);
  lseek (devd, 0, SEEK_SET);
  if (write (devd, buf_ldr, SECTOR_SIZE) != SECTOR_SIZE)
    die ("write %s: %s", ldr_dev, strerror (errno));
  close_blk_dev (&devd);

  devd = open_blk_dev (knl_dev);
  lseek (devd, SECTOR_SIZE, SEEK_SET);
  if (write (devd, buf_knl, ks) != ks)
    die ("write %s: %s", knl_dev, strerror (errno));
  close_blk_dev (&devd);

#else

  if (fread (&ldev_id, sizeof (ldev_id), 1, ui_fp) < 1 ||
      fread (buf_ldr, SECTOR_SIZE, 1, ui_fp) < 1 ||
      fread (&kdev_id, sizeof (kdev_id), 1, ui_fp) < 1 ||
      fread (&ks, sizeof (ks), 1, ui_fp) < 1 ||
      fread (buf_knl, ks * 512, 1, ui_fp) < 1)
    die ("fread backup file error");

  check_dev (ldev_id);
  disk_io (3, ldev_id, LDR_HEAD, LDR_CYL, LDR_SECT, buf_ldr, 1);

  check_dev (kdev_id);
  disk_io (3, kdev_id, KNL_HEAD, KNL_CYL, KNL_SECT, buf_knl, ks);

#endif

  fclose(ui_fp);
}

void
check_lilo ()
{
  int devd;
  char buf[512];

  /* searching lilo */

#ifdef __linux__

  devd = open_blk_dev (ldr_dev);

  if (read (devd, buf, 10) != 10)
    die ("read %s: %s", ldr_dev, strerror (errno));

  close_blk_dev (&devd);

#else

  disk_io (2, ldev_id, LDR_HEAD, LDR_CYL, LDR_SECT, buf, 1);

#endif

  buf[10] = '\0';
  if (strcmp (buf + 6, "LILO") == 0)
    {
#ifdef __linux__
      printf ("It seems that LILO has been installed into device %s\n"
              "Do you want to continue to install SmartBootManager [yes/no]?",
             ldr_dev);
#else
      printf ("It seems that LILO has been installed into drive %d\n"
              "Do you want to continue to install SmartBootManager [yes/no]?",
             ldev_id);
#endif

      fgets (buf, 5, stdin);
      if (strncmp (buf, "yes", 3) != 0)
    die ("Abort installation.");
    }
}

int
main (int argc, char *argv[])
{
  char input[20];

  printf (about);

  init_settings ();
  grasp_params (argc, argv);

  if (ui_fp != NULL)
    {
      unistall ();
      printf ("Uninstallation successful!");
      exit (0);
    }

  p_theme = (struct sbmt_header *) theme_code;
  p_knl = (struct sbmk_header *) main_code;
  p_loader = (struct sbml_header *) buf_ldr;

  if (p_theme->magic != SBMT_MAGIC)
    die ("Invalid theme file!");
  if (p_theme->version != p_knl->version)
    die ("Theme version mismatched!");
  knl_tsize = p_knl->kernel_size + p_theme->size;
  if (p_theme->size > MAX_THEME_SIZE)
    die ("Theme file is too big!");

  if (bk_fp != NULL)
    backup ();

  if (!always_yes)
    {
      check_lilo ();

#ifdef __linux__
      printf ("Are you sure to install Smart BootManager into drive %s [y/n]?",
	      ldr_dev);
#else
      printf
        ("Are you sure to install Smart BootManager into drive %d [y/n]?",
         ldev_id);
#endif

      fgets (input, 19, stdin);

      if (input[0] != 'y')
	die ("Abort installation.");
    }

  proc_knl ();
  proc_ldr ();

  printf ("\nInstallation successful!\n");
  exit (0);
}
