; install.asm
;
; Simple installation program for Smart Boot Manager 3.0
;
; Copyright (C) 2000, Suzhe. See file COPYING for details.
;
%define MAIN

%include "ui.h"
%include "hd_io.h"
%include "sbm.h"

%define DRVID        0x80
%define LILO_ID_OFF  6
%define LILO_MAGIC   'LILO'

        org 0x100
        section .text

main:
        mov ax, cs
        mov ds, ax
        mov es, ax
        
        call init_video

        lea si, [strings.confirm]
        call info_box
        cmp al, 'y'
        je .inst_sbm
        cmp al, 'Y'
        je .inst_sbm
        lea si, [strings.abort]
        call error_box
        jmp .exit
.inst_sbm:
        call install_sbm
.exit:
        call reset_video_mode
        mov ax, 0x4c00
        int 0x21

;=============================================================================
;init_video ---- init the video mode
;input:
;      al = video mode ( 0: 90x25, otherwise: 80x25 )
;output:
;      none
;=============================================================================
init_video:
        mov al, 0xff
        mov bl, 8
        xor bp, bp
        xor cx, cx
;        lea bp, [font.data]
;        mov cx, [font.number]
        call set_video_mode
        ret

install_sbm:
; read boot sector
        mov ax, 0x0201
        xor cx, cx
        xor bx, bx
        mov dl, DRVID
        lea di, [disk_buf]
        call disk_access
        jnc .check_lilo
        jmp .disk_err
        
.check_lilo:
; check lilo
        mov si, di
        mov eax, [si + LILO_ID_OFF]
        cmp eax, LILO_MAGIC
        jne .copy_pt
        jmp .lilo_present

.copy_pt:
; copy partition table
        add si, PART_TBL_OFF
        lea di, [loader_data]
        add di, PART_TBL_OFF
        mov cx, 64
        cld
        rep movsb

        xor eax, eax
        
; copy theme data
        lea di, [kernel_data]
        lea si, [theme_data]
        push di
        mov ax, [di + struc_sbmk_header.kernel_size]
        add di, ax
        mov cx, [si + struc_sbmt_header.size]
        add ax, cx
        rep movsb
        pop di

; calculate kernel sectors
        add ax, 511
        xor dx, dx
        mov cx, 512
        div cx
        
; fill in the kernel_sectors
        mov [di + struc_sbmk_header.kernel_sectors], al
        lea si, [loader_data]
        mov [si + struc_sbml_header.kernel_sectors], al

; fill in the loader header
        mov byte [si + struc_sbml_header.kernel_copies], 2
        mov byte [si + struc_sbml_header.kernel_addr], DRVID
        mov dword [si + struc_sbml_header.kernel_addr + 1], 1
        mov byte [si + struc_sbml_header.kernel_addr + 5], DRVID
        mov dword [si + struc_sbml_header.kernel_addr + 6], 1
        add dword [si + struc_sbml_header.kernel_addr + 6], eax

; fill in the kernel header
        mov byte [di + struc_sbmk_header.kernel_copies], 1
        mov byte [di + struc_sbmk_header.kernel_addr], DRVID
        mov dword [di + struc_sbmk_header.kernel_addr + 1], 1

; calculate the checksum
        push eax
        mov si, di
        call calc_checksum
        neg bl
        mov [di + struc_sbmk_header.checksum], bl

; write the first kernel
        mov ah, 0x03
        xor cx, cx
        mov bx, 1
        mov dl, DRVID
        call disk_access
        pop eax
        jc .disk_err
        
; write the second kernel
        add dword [di + struc_sbmk_header.kernel_addr + 1], eax

; calculate the checksum
        push ax
        mov si, di
        call calc_checksum
        neg bl
        mov [di + struc_sbmk_header.checksum], bl
        pop ax
        
        mov ah, 0x03
        xor cx, cx
        xor bh, bh
        mov bl, al
        inc bl
        mov dl, DRVID
        call disk_access
        jc .disk_err

; write loader
        lea di, [loader_data]
        mov ax, 0x0301
        xor cx, cx
        xor bx, bx
        mov dl, DRVID
        call disk_access
        jc .disk_err

        lea si, [strings.install_ok]
        call info_box
        ret

.disk_err:
        lea si, [strings.disk_err]
        jmp .report_err
.lilo_present:
        lea si, [strings.lilo_present]
.report_err:
        call error_box
        ret

;=============================================================================
;calc_checksum ---- calculate the checksum of the kernel program.
;input:
;      si -> kernel data
;output:
;      bl = the checksum value.
;=============================================================================
calc_checksum:
        push ax
        push cx
        push si
        mov cx, [si + struc_sbmk_header.checksum_size]
        xor bl, bl
        mov [si + struc_sbmk_header.checksum], bl

.loop_calc:
        lodsb
        add bl, al
        loop .loop_calc
        pop si
        pop cx
        pop ax
        ret

;=============================================================================
;error_box ---- draw error message box.
;input:
;      ds:si -> error message
;output:
;      none
;=============================================================================
error_box:
        call strlen
        mov dl, cl
        add dl, 4
        mov dh, [size.box_height]
        mov al, [color.error_box_msg]
        mov bx, [color.error_box_frame]
        lea di, [strings.error]
        xchg di, si
        call message_box
        xor ah, ah
        int 0x16
        ret

;=============================================================================
;info_box ---- draw infomation message box.
;input:
;      ds:si -> infomation message
;output:
;      none
;=============================================================================
info_box:
        call strlen
        mov dl, cl
        add dl, 4
        mov dh, [size.box_height]
        mov al, [color.info_box_msg]
        mov bx, [color.info_box_frame]
        lea di, [strings.info]
        xchg di, si
        call message_box
        xor ah, ah
        int 0x16
        ret


;===========================================================================
; DATA AREA
;===========================================================================

%include "ui.asm"
%include "utils.asm"
%include "hd_io.asm"

color:
.background         db  0x00        ; background (if no background icon)
.error_box_frame    db  0xCF        ;
.error_box_title    db  0xF1        ; error box
.error_box_msg      db  0xCF        ;
.info_box_frame     db  0xB0        ;
.info_box_title     db  0xF1        ; info box
.info_box_msg       db  0xB0        ;

frame_char:
.top                db     0x20            ; top horizontal
.bottom             db     0xCD            ; bottom horiztontal
.left               db     0xBA            ; left vertical
.right              db     0xBA            ; right vertical
.tl_corner          db     0xC9            ; top left corner
.tr_corner          db     0xBB            ; top right corner
.bl_corner          db     0xC8            ; bottom left corner
.br_corner          db     0xBC            ; bottom right corner

size:
.box_height      db  5

str_idx:
.input          dw  strings.input

loader_data:
        incbin "loader.bin"
kernel_data:
        incbin "main.bin"
theme_data:
        incbin "theme-zh.bin"
        
strings:
.input         db "Input",0
.info          db "Information",0
.error         db "Error",0
.confirm       db "You are about to install SmartBootManager, Press y to continue.",0
.abort         db "Abort the installation.",0
.install_ok    db "Installation Success.",0
.disk_err      db "Disk read/write error!",0
.lilo_present  db "Found LILO in MBR!",0

; how to draw window frame
draw_frame_method   db  1          ; = 0 means draw all frame using frame attr.
                                   ; = 1 means draw top horizontal line using
                                   ;     title attr.
                                   ; = 2 means draw top corner and horizontal
                                   ;     line using title attr.

%include "tempdata.asm"

disk_buf    times 512 db 0

