; main.asm
;
; Main programs for Smart Boot Manager
;
; Copyright (C) 2000, Suzhe. See file COPYING for details.
;
%define MAIN
%define HAVE_MAIN_PROG
;%define EMULATE_PROG

%include "ui.h"
%include "hd_io.h"
%include "knl.h"
%include "sbm.h"

%define MAIN_WIN_WIDTH      52
%define MENU_POS_ROW        3
%define MENU_POS_COL        2
%define MENU_TITLE_POS_ROW  2
%define MENU_TITLE_POS_COL  2
%define MENU_WIDTH          47
%define DELAY_TIME_ROW      11
%define DELAY_TIME_COL      37

%define MAX_RECORD_NUM      32
%define MAX_PASSWORD_LENGTH 16

%define NUM_OF_HOTKEYS      (func_list.entries - func_list.hotkeys)/2

%define KNLFLAG_FIRSTRUN    0x01
%define KNLFLAG_SECURITY    0x02
%define KNLFLAG_FLOPPY_OK   0x04

        bits 16

%ifdef EMULATE_PROG
        org 0x100
%else
        org 0
%endif

        section .text
        
start_of_kernel:
start_of_checksum_area:

        jmp start
        nop
;!!! PLEASE DON NOT CHANGE THE SIZE AND ORDER OF FOLLOWING DATA !!!

;=============================================================================
;the header of Smart Boot Manager kernel ( 8 bytes )
;=============================================================================
sbmk_magic          dd  SBMK_MAGIC ; magic number = 'SBMK', 4 bytes.
                                   ; it's abbr. of 'Smart Boot Manager Kernel'
sbmk_version        dw  SBMK_VERSION ; version, high byte is major version,
                                   ; low byte is minor version.
kernel_size         dw  (end_of_kernel-start_of_kernel)
checksum_size       dw  (end_of_checksum_area - start_of_checksum_area)
checksum            db  0          ; checksum of the kernel program.

kernel_sectors      db  0          ; the number of sectors that kernel and theme
                                   ; totally used.
kernel_copies       db  0          ; number of kernel copies saved. <= 3
kernel_drvid        db  0          ; the drive id that kernel saved.
kernel_addr         dd  0          ; the abs sector addr that kernel saved.

kernel_drvid2       db  0          ;
kernel_addr2        dd  0          ;

kernel_drvid3       db  0          ;
kernel_addr3        dd  0          ;

;=============================================================================
; MAIN data area
;=============================================================================
kernel_flags        db 0x05                     ; first run signature 
                                                ; and floppy ok signature.
delay_time          db 30                       ; delay time ( seconds )
direct_boot_record  db 0xff                     ; >= MAX_RECORD_NUM means no
                                                ; direct boot.
default_record      db 0xff                     ; the record number will
                                                ; be booted after the
                                                ; delay time is up or ESC
                                                ; key is pressed.
root_password       dd 0

; buffer to store boot records
boot_records        times MAX_RECORD_NUM * SIZE_OF_BOOTRECORD db 0
;=============================================================================

start:

%ifndef EMULATE_PROG                            ; real program goes here.

        cli
        mov ax, cs                              ; initialize the segment
        mov ds, ax                              ; registers and the stack
        mov es, ax                              ; pointer.
        xor ax, ax
        mov ss, ax                              ; ss:sp = 0:0xA000
        mov sp, KERNEL_OFF                      ;
        sti
        
        test byte [kernel_flags], KNLFLAG_FIRSTRUN
        jz .not_first_run
        
        call init_theme                         ; initialize the theme, only time!
        call init_video                         ; initialize the video mode.
        call init_boot_records                  ; if it's the first time
        and byte [kernel_flags], ~ KNLFLAG_FIRSTRUN ; to run this program,
        call save_boot_manager                  ; call the init_boot_records.
        jc .disk_error
        jmp .show_menu
        
.not_first_run:
        
        call init_video                         ; initialize the video mode.

        mov ah, 0x02                            ; test the keyboard status,
        int 0x16                                ; if ctrl pressed then show
        test al, kbCtrlMask                     ; menu directly,
        jnz .show_menu                          ;
        
        mov bl, SIZE_OF_BOOTRECORD

        mov al, [direct_boot_record]            ; check if need boot directly.
        cmp al, MAX_RECORD_NUM                  ;
        jb .go_direct_boot

        call do_schedule                        ; implement the schedule table.

        cmp byte [delay_time], 0
        jnz .show_menu                          ; delay_time = 0, boot the
                                                ; default record directly.
        mov al, [default_record]                ;
        cmp al, MAX_RECORD_NUM
        jb .go_def_boot
        jmp .show_menu
        
.go_direct_boot:
        mov byte [direct_boot_record], 0xff     ; clear the direct boot sig.
        call save_boot_manager
        jc .disk_error

.go_def_boot:
        mul bl                                  ; get the pointer to the
        lea si, [boot_records]                  ; direct/default boot record.
        add si, ax                              ;

        call check_bootrecord                   ; check if it's valid
        jc .invalid_record

        call boot_the_record                    ; boot it.
        
        or al, al
        jnc .disk_error
        mov si, [str_idx.no_system]
        jmp .report_error
        
.disk_error:
        mov si, [str_idx.disk_error]
        jmp .report_error
.invalid_record:
        mov si, [str_idx.invalid_record]
.report_error:
        call error_box

%else
        call init_video                         ; here is the code for
        call init_boot_records                  ; emulate program.
        call do_schedule
%endif

.show_menu:
        xor al, al
        mov [change_occured], al
        mov [root_login], al
        
        mov al, [delay_time]                    ;
        cmp al, 255                             ;
        jae .not_count_time                     ;
                                                ; if delay_time = 255
        mov [time_count], al                    ; then do not count time.
        xor al, al                              ;
.not_count_time:                                ;
        mov [key_pressed], al                   ;
        
        xor al, al
        mov [first_visible_rec], al
        mov [focus_record], al
        
        call init_good_record_list

        mov bl, [default_record]
        lea si, [good_record_list]
        mov cl, [good_record_number]
        xor ch, ch
        xor bh, bh
        
        cld
.loop_search_def:
        lodsb
        cmp al, bl
        je .found_def
        inc bh
        loop .loop_search_def
        jmp .go_ahead

.found_def:
        mov [focus_record], bh
        cmp bh, [size.bootmenu_height]
        jb .go_ahead
        inc bh
        sub bh, [size.bootmenu_height]
        mov [first_visible_rec], bh

.go_ahead:
        call draw_screen
        call msg_dispatcher                     ; go to message dispatcher
                                                ; this function may not return.

;=============================================================================
;init_theme ---- initialize the theme data.
;=============================================================================
init_theme:
        mov bx, [icon.brand]
        or bx, bx
        jz .adjust_bkgnd                            ; no brand icon
        add word [icon.brand], theme_start          ;
.adjust_bkgnd:
        mov bx, [icon.background]
        or bx, bx
        jz .adjust_font                             ; no background icon
        add word [icon.background], theme_start
.adjust_font:
        mov bx, [font.data]
        or bx, bx
        jz .adjust_str
        add word [font.data], theme_start           ; no font data

.adjust_str:
        lea si, [str_idx]
        mov cx, (end_of_str_idx - str_idx)/2
        
.loop_adjust:
        mov bx, [si]
        add bx, theme_start
        mov [si], bx
        add si, 2
        loop .loop_adjust

        mov al, 0x10
        and [keyboard_type], al

        ret
        
;=============================================================================
;init_video ---- init the video mode
;input:
;      al = video mode ( 0: 90x25, otherwise: 80x25 )
;output:
;      none
;=============================================================================
init_video:
        mov al, [video_mode]
        mov bl, 8
        mov bp, [font.data]
        mov cx, [font.number]
        
        call set_video_mode
        ret

;=============================================================================
;init_good_record_list ---- init the good boot record list
;input:
;      none
;output:
;      cf = 0 sucess
;      cf = 1 failed, no good record
;=============================================================================
init_good_record_list:
        pusha
        mov cx, MAX_RECORD_NUM
        lea di, [good_record_list]
        lea si, [boot_records]
        xor ax, ax
.loop_check:
        call check_bootrecord                   ; check if it's valid
        jc .check_next

        cld
        stosb                                   ; store it's index to buffer
        inc ah
        
.check_next:
        inc al
        add si, SIZE_OF_BOOTRECORD
        loop .loop_check

        mov [good_record_number], ah
        or ah, ah
        jnz .ok
        stc
.ok:
        popa
        ret

;=============================================================================
;init_boot_records ---- init the boot records list
;input:
;      none
;output:
;      cf = 0 success
;      cf = 1 failed, no good record
;=============================================================================
init_boot_records:
        pusha
        inc byte [change_occured]               ; some changes occured.

        cld
        mov al, SIZE_OF_BOOTRECORD
        mov cl, MAX_RECORD_NUM
        mul cl
        push cx
        mov cx, ax
        lea si, [boot_records]
        lea di, [tmp_records_buf]
        push cx
        push di
        push si
        rep movsb                              ; backup current bootrecords
        pop di                                 ; di = [boot_records]
        pop si                                 ; si = [tmp_records_buf]
        pop cx
        push di
        xor al, al                             ;
        rep stosb                              ; clear boot_records 
        pop di
        pop cx
        
        xor dl, dl
        test byte [kernel_flags], KNLFLAG_FLOPPY_OK
        jnz .search_it
        mov dl, MIN_HD_ID
.search_it:
        call search_drives
        or ch, ch
        jnz .ok
        stc
        popa
        ret
.ok:
; compare old records with new, find out the unchanged records.
        mov cx, MAX_RECORD_NUM
.compare_new_next:
        push cx
        test byte [di + struc_bootrecord.flags], INFOFLAG_DRIVEOK
        jz .skip_new

        mov cx, MAX_RECORD_NUM
        lea si, [tmp_records_buf]

.compare_old_next:
        test byte [si + struc_bootrecord.flags], INFOFLAG_DRIVEOK
        jz .not_same

        mov ax, [di + struc_bootrecord.drive_id]
        cmp [si + struc_bootrecord.drive_id], ax
        jne .not_same
        mov al, [di + struc_bootrecord.type]
        cmp [si + struc_bootrecord.type], al
        jne .not_same
        mov eax, [di + struc_bootrecord.father_abs_addr]
        cmp [si + struc_bootrecord.father_abs_addr], eax
        jne .not_same
        mov eax, [di + struc_bootrecord.abs_addr]
        cmp [si + struc_bootrecord.abs_addr], eax
        jne .not_same

; found same boot records, copy some information!
        mov al, [si + struc_bootrecord.flags]
        and al, INFOFLAG_AUTOHIDE + INFOFLAG_AUTOACTIVE
        or [di + struc_bootrecord.flags], al

        push si
        push di
        push cx
        mov cx, struc_bootrecord.end_of_struc - struc_bootrecord.password
        mov ax, struc_bootrecord.password
        add di, ax
        add si, ax
        cld
        rep movsb
        pop cx
        pop di
        pop si
        jmp .skip_new

.not_same:
        add si, SIZE_OF_BOOTRECORD
        loop .compare_old_next

.skip_new:
        pop cx
        add di, SIZE_OF_BOOTRECORD
        loop .compare_new_next

        popa
        ret

;=============================================================================
;draw_screen ---- draw the whole screen
;input:
;      none
;output:
;      none
;=============================================================================
draw_screen:
        mov bh, [color.background]              ;
        mov si, [icon.background]               ; draw background
        mov cx, [icon.background_size]          ;
        call draw_background                    ;

        xor dx, dx                              ;
        mov bl, [color.copyright]               ;
        mov al, [screen_width]                  ; draw copyright message
        push ax                                 ; save screen width
        mov cl, [size.copyright]
        mul cl
        mov cx, ax
        mov al, ' '                             ;
        call draw_char                          ;
        mov si, [str_idx.copyright]
        call draw_string
        
        mov bl, [color.hint]                    ;
        mov dh, [screen_height]                 ;
        mov cl, [size.hint]                     ;
        sub dh, cl                              ; draw hint message
        pop ax                                  ; get screen width
        mul cl                                  ;
        mov cx, ax                              ;
        mov al, ' '                             ;
        call draw_char                          ;
        mov si, [str_idx.hint]                  ;
        call draw_string                        ;

        mov dx, [position.brand_col]            ; draw brand icon
        mov cx, [icon.brand_size]               ;
        cmp dl, 0xFF                            ;
        jne .not_justify                        ;
        mov dl, [screen_width]                  ; right justify
        sub dl, cl                              ;
.not_justify:                                   ;
        mov si, [icon.brand]                    ;
        call draw_icon                          ;

        call draw_delay_time
        call draw_security_flags
        call draw_main_window
        call draw_bootmenu
        ret
        
;=============================================================================
;draw_main_window ---- draw the main window
;input:
;      none
;output:
;      none
;=============================================================================
draw_main_window:
        pusha
        mov cx, [position.main_win_col]                 ; get main window's
        mov dx, cx                                      ; coordinate
        add dh, [size.main_win_height]
        dec dh
        add dl, (MAIN_WIN_WIDTH-1)

        mov bx, [color.main_win_frame]                  ; get main window's color
        mov si, [str_idx.main_win_title]                ; get main window's title

        call draw_window

        add cx, (MENU_TITLE_POS_ROW << 8) + MENU_TITLE_POS_COL
        mov dx, cx
        mov bl, [color.menu_title]
        mov al, ' '
        mov cx, MENU_WIDTH
        call draw_char
        mov si, [str_idx.menu_title]
        call draw_string

        popa
        ret

;=============================================================================
;draw_security_flags ---- draw root passwd, login and secure mode flags.
;=============================================================================
draw_security_flags:
        mov dh, [screen_height]
        mov dl, [screen_width]
        mov bl, [color.security_flags]
        sub dx, 0x010d
        mov cx, 1

        cmp dword [root_password], 0
        jz .no_root_password
        
        mov al, 'P'
        jmp .draw_pwd
.no_root_password:
        mov al, ' '
.draw_pwd:
        call draw_char
        inc dl

        test byte [kernel_flags], KNLFLAG_SECURITY
        jz .no_security

        mov al, 'S'
        jmp .draw_security
.no_security:
        mov al, ' '
.draw_security:
        call draw_char
        inc dl

        mov al, [root_login]
        or al, al
        jz .no_root_login

        mov al, 'A'
        jmp .draw_login
.no_root_login:
        mov al, ' '
.draw_login:
        call draw_char
        ret

;=============================================================================
;draw_scrollbar ---- draw the scroll bar
;=============================================================================
draw_scrollbar:
        xor cx, cx
        inc cl

        mov bl, [color.scrollbar]
        mov dx, [position.main_win_col]
        add dx, (MENU_POS_ROW << 8) + MENU_POS_COL
        add dl, MENU_WIDTH

        mov bh, [size.bootmenu_height]
        push dx
        push bx

;draw scroll bar
        mov al, 0x1e
        call draw_char

        mov al, ' '
        dec bh
.loop_draw_bar:
        dec bh
        inc dh
        call draw_char
        or bh, bh
        jnz .loop_draw_bar

        mov al, 0x1f
        call draw_char

;calculate the bar cursor position
        pop bx
        pop dx

        cmp bh, [good_record_number]
        jae .no_cursor
        cmp bh, 3
        jb .no_cursor

        sub bh, 2

        mov al, [focus_record]
        xor ah, ah
        mul bh
        mov bh, [good_record_number]
        or bh, bh
        jz .no_cursor
        div bh

        add dh, al
        inc dh

        mov al, 'O'
        call draw_char

.no_cursor:
        ret
 
;=============================================================================
;draw_bootmenu_item ---- draw a boot menu item
;input:
;      bl = high 4 bit Background color and low 4 bit Foreground color
;      cl = item index in good_record_list
;output:
;      none
;=============================================================================
draw_bootmenu_item:
        pusha
        cmp byte [good_record_number], 0
        je .no_menu_item

        xor ch, ch
        lea si, [good_record_list]                  ; get real index of the
        add si, cx                                  ; boot record.
        lodsb                                       ; al = the index
        
        xor ah, ah                                  ; get the pointer of
        push ax                                     ;
        mov dl, SIZE_OF_BOOTRECORD                  ; selected boot record.
        mul dl                                      ;
        lea si, [boot_records]                      ;
        add si, ax                                  ;
        
        lea di, [tmp_buffer]
        call get_record_string

        mov dh, cl                                  ; calculate the draw
        sub dh, [first_visible_rec]                 ; coordinate of the
        add dh, MENU_POS_ROW                        ; menu item.
        mov dl, MENU_POS_COL                        ;
        add dx, [position.main_win_col]             ;

        mov cl, 1
        mov al, ' '
        call draw_char                              ; draw a leading space
        inc dl
        pop ax                                      ; al = the real index

        cmp al, [default_record]                    ;
        jne .not_default                            ;
        mov al, '*'                                 ; draw a flag before
        jmp .draw_def_flag                          ; the menu item, if
.not_default:                                       ; it's default record.
        mov al, ' '                                 ;
.draw_def_flag:                                     ;
        call draw_char                              ;
        inc dl                                      ;
        
        mov si, di
        call draw_string                            ; draw the menu item.

        call strlen                                 ; get the length of the item.
        mov al, ' '
        mov ah, MENU_WIDTH
        sub ah, cl
        sub ah, 3
        add dl, cl
        mov cl, ah
        inc cl
        call draw_char                              ; fill ending space chars
.no_menu_item:
        popa
        ret

;=============================================================================
;draw_bootmenu ---- draw the boot menu
;input:
;      none
;output:
;      none
;=============================================================================
draw_bootmenu:
        pusha
        mov bl, [color.menu_normal]
        mov bh, bl
        mov cx, [position.main_win_col]                 ; get the top left
        add cx, (MENU_POS_ROW << 8) + MENU_POS_COL      ; and bottom right
        mov dx, cx                                      ; corner's coordinate
        add dh, [size.bootmenu_height]
        dec dh
        add dl, (MENU_WIDTH-1)                          ; of menu area

        mov ax,0x0600                                   ; clear the menu area
        int 0x10                                        ;

        mov cl, [first_visible_rec]
        mov ch, cl
        add ch, [size.bootmenu_height]
.loop_draw:
        cmp cl, [good_record_number]
        jae .end
        cmp cl, ch
        jae .end
        call draw_bootmenu_item
        inc cl
        jmp .loop_draw
.end:

;draw focused item
        mov bl, [color.menu_focus]
        mov cl, [focus_record]

        call draw_bootmenu_item                     ; draw focus menu item
        call draw_scrollbar

        popa
        ret

;=============================================================================
;get_key ---- get a key stroke
;input:
;      none
;output:
;      ax = the key code
;=============================================================================
get_key:

.wait_key:
        cmp byte [delay_time], 0                        ; if delay_time set to zero
        je .no_count                                    ; then don't count time.
        cmp byte [key_pressed], 0                       ; if key pressed then don't
        jne .no_count                                   ; count time.

        xor ah, ah                                      ; get time ticks
        int 0x1a                                        ;

        cmp dx, [ticks_count]                           ;
        jae .next_time                                  ; dx must greater than
        mov [ticks_count], dx                           ; ticks_count
.next_time:
        mov cx, dx                                      ; every 18 ticks approxmiately
        sub cx, [ticks_count]                           ; equal to 1 second,
        cmp cx, 18                                      ; decrease time_count
        jbe .not_add                                    ; until to zero.
        mov [ticks_count], dx                           ;
        dec byte [time_count]                           ;
        call draw_delay_time                            ;
.not_add:
        cmp byte [time_count], 0                        ; if time is up, then
        jbe .esc_key                                    ; send ESC key.

.no_count:
        mov ah,1                                        ; if no key pressed
        or ah, [keyboard_type]
        int 0x16                                        ; go back to check
        jz .wait_key                                    ; time.
        mov ah, [keyboard_type]                         ; if any key pressed
        int 0x16                                        ; then get it.
        mov byte [key_pressed], 1
        ret
.esc_key:
        mov ax, kbEsc
        inc byte [key_pressed]
        ret

;=============================================================================
;draw_delay_time ---- draw the delay_time and time_count
;=============================================================================
draw_delay_time:
        mov al, [time_count]
        mov cx, 3
        xor ah, ah
        lea di, [tmp_buffer]
        call itoa
        
        mov bl, [color.delay_time]
        mov dh, [screen_height]
        mov dl, [screen_width]
        sub dx, 0x0108
        mov si, di
        call draw_string

        mov al, [delay_time]
        mov cx, 3
        xor ah, ah
        call itoa

        mov al, '|'
        mov cl, 1
        add dl, 3
        call draw_char
        inc dl
        call draw_string
        mov al, ' '
        add dl, 3
        call draw_char
        ret
        
        
;=============================================================================
;msg_dispatcher ---- the main loop to handle user command
;input:
;      none
;output:
;      none
;=============================================================================
msg_dispatcher:
.loop_dispatch:
        call draw_bootmenu
;        mov bl, [color.menu_focus]
;        mov cl, [focus_record]

;        call draw_bootmenu_item                     ; draw focus menu item
;        push cx
        call get_key                            ; get a key
;        pop cx
;        mov bl, [color.menu_normal]
;        call  draw_bootmenu_item                    ; clear focus menu item

        lea si, [func_list.hotkeys]             ; search in hotkey list
        mov dx, ax                              ;
        mov cx, NUM_OF_HOTKEYS                  ;
        xor bx, bx                              ;
        cld
.loop_search:                                   ;
        lodsw                                   ;
        cmp dx, ax                              ;
        je .found_it                            ; found, go get func entry
        inc  bl                                 ;
        loop .loop_search                       ;
        jmp .loop_dispatch                      ; not found, wait more

.found_it:
        shl bl, 1
        call word [bx + func_list.entries]      ; call the function
        jmp .loop_dispatch


;=============================================================================
;below are key stroke handlers
;=============================================================================

;=============================================================================
;show_help ---- show the help window
;=============================================================================
show_help:
        mov di, [str_idx.help]
        mov si, [str_idx.help_content]
        or si, si
        jz .end
        
        call count_lines
        mov dx, cx
        add dh, [size.box_height]
        add dl, [size.box_width]
        mov al, [color.help_msg]
        mov bx, [color.help_win_frame]
        xchg si, di
        call message_box
        xor ah, ah
        int 0x16
        call draw_screen
.end:
        ret

;=============================================================================
;show_about ---- show the about window
;=============================================================================
show_about:
        mov di, [str_idx.about]
        mov si, [str_idx.about_content]
        or si, si
        jz .end
        
        call count_lines
        mov dx, cx
        add dh, [size.box_height]
        add dl, [size.box_width]
        mov al, [color.about_msg]
        mov bx, [color.about_win_frame]
        xchg si, di
        call message_box
        xor ah, ah
        int 0x16
        call draw_screen
.end:
        ret

;=============================================================================
;calc_checksum ---- calculate the checksum of the kernel program.
;input:
;      none
;output:
;      bl = the checksum value.
;=============================================================================
calc_checksum:
        lea si, [start_of_checksum_area]
        mov cx, [checksum_size]
        xor bl, bl
        mov [checksum], bl
        cld
.loop_calc:
        lodsb
        add bl, al
        loop .loop_calc
        
        ret
        
;=============================================================================
;save_boot_manager ---- save boot manager to disk.
;input:
;      none
;output:
;      cf = 0 success
;      cf = 1 failed
;=============================================================================
save_boot_manager:
        pusha
        call calc_checksum                      ; calculate the checksum.
        neg bl
        mov [checksum], bl                      ;

        lea si, [kernel_drvid]
        xor ch, ch
        mov cl, [kernel_copies]
        cld
.loop_save:
        push cx
        
        lodsb
        mov dl, al
        lodsw
        mov bx, ax
        lodsw
        mov cx, ax
        
        mov dh, [kernel_sectors]
        mov ah, INT13H_WRITE
        mov al, 1
        xor di, di

.loop_write:
        clc
        or dh, dh
        jz .write_end
        call disk_access

        jc .write_end
        
        add di, SECTOR_SIZE
        adc bx, 1                      ; inc cx : bx, read next sector.
        adc cx, 0                      ;
        dec dh
        jmp .loop_write
        
.write_end:

        pop cx
        
        jc .disk_error
        loop .loop_save

        mov byte [change_occured], 0               ; clear change signature.
        
        clc
.disk_error:
        popa
        ret
        
;=============================================================================
;ask_save_changes ---- save boot manager to disk
;=============================================================================
ask_save_changes:
        cmp byte [change_occured], 0
        je .no_changes
        
        mov si, [str_idx.ask_save_changes]
        call info_box
        cmp ax, kbEnter
        je save_changes
        cmp ax, kbLowerY
        je save_changes
        cmp ax, kbUpperY
        je save_changes
        
.no_changes:
        ret

;=============================================================================
;save_changes ---- save boot manager to disk
;=============================================================================
save_changes:
        test byte [kernel_flags], KNLFLAG_SECURITY
        jz .normal
        call confirm_root_passwd
        jc .end

.normal:

%ifndef EMULATE_PROG
        call save_boot_manager
        jc .disk_error
%endif

        mov si, [str_idx.changes_saved]
        call info_box
        ret
        
.disk_error:
        mov si, [str_idx.disk_error]
        call error_box
.end:
        ret

;=============================================================================
;change_video_mode ---- change the video mode
;=============================================================================
change_video_mode:

        inc byte [change_occured]               ; some changes occured.

        mov al, [video_mode]
        not al
        mov [video_mode], al
        call init_video
        call draw_screen
        ret
        
;=============================================================================
;up_cursor ---- move the cursor up
;=============================================================================
up_cursor:
        mov al, [focus_record]
        or al, al                               ; already at top of the list
        jz .end                                 ; do nothing
        dec al
        mov [focus_record], al
        cmp [first_visible_rec], al
        jbe .end                                ; no need to scroll down.
        mov [first_visible_rec], al             ;
.end:
        ret
        
;=============================================================================
;down_cursor ---- move the cursor down
;=============================================================================
down_cursor:
        mov al, [focus_record]
        inc al
        cmp al, [good_record_number]            ;
        jae .end                                ; already at bottom of the list
        mov [focus_record], al
        sub al, [first_visible_rec]
        cmp al, [size.bootmenu_height]
        jb .end                                 ; no need to scroll up
        inc byte [first_visible_rec]
.end:
        ret

;=============================================================================
;change_name ---- change the record name
;=============================================================================
change_name:
        cmp byte [good_record_number], 0
        je .end
        
        call confirm_security_passwd
        jc .end

        call get_record_pointer

        add si, struc_bootrecord.name                   ;
        lea di, [tmp_buffer]                            ; get old name.
        mov cl, MAX_NAME_LENGTH                         ;
        xor ch, ch                                      ;
        push cx                                         ;
        push di                                         ;
        call strncpy                                    ;
        pop di                                          ;
        pop cx                                          ;
        
        push si
        
        xor ah, ah
        mov al, [color.input_box_msg]
        mov bx, [color.input_box_frame]
        mov si, [str_idx.name]
        call input_box
        pop si
        jc .end_clean


        xchg di, si
        xor ch, ch
        call strncpy

        inc byte [change_occured]               ; some changes occured.
        
.end_clean:
        call draw_screen
.end:
        ret


;=============================================================================
;login_as_root ---- login as root
;=============================================================================
login_as_root:
        cmp dword [root_password], 0
        jz .end

        mov al, [root_login]
        or al, al
        jnz .logout

        call confirm_root_passwd
        jc .end

        mov byte [root_login], 1
        jmp .end_clean

.logout:
        xor al, al
        mov [root_login], al

.end_clean:
        call draw_security_flags
.end:
        ret


;=============================================================================
;change_security_mode ---- change the secure mode
;=============================================================================
change_security_mode:
        cmp dword[root_password], 0
        jz .end

        test byte [kernel_flags], KNLFLAG_SECURITY
        jz .enter_security

        call confirm_root_passwd
        jc .end

        and byte [kernel_flags], ~ KNLFLAG_SECURITY
        jmp .end_clean

.enter_security:
        or byte [kernel_flags], KNLFLAG_SECURITY

.end_clean:
        inc byte [change_occured]
        call draw_security_flags
.end:
        ret

;=============================================================================
;change_root_passwd ---- change the root password
;=============================================================================
change_root_passwd:
        call confirm_root_passwd
        jc .end
        
;        call draw_screen
        
        mov si, [str_idx.new_root_passwd]
        call input_passwd
        jc .end
        mov bx, ax
        mov cx, dx

        push bx
        push cx
        mov si, [str_idx.retype_passwd]
        call input_passwd
        pop cx
        pop bx
        jc .end
        cmp bx, ax
        jne .wrong
        cmp cx, dx
        jne .wrong
        mov [root_password], bx
        mov [root_password+2], cx

        mov byte [root_login], 0
        and byte [kernel_flags], ~ KNLFLAG_SECURITY

        mov si, [str_idx.passwd_changed]
        call info_box

        inc byte [change_occured]               ; some changes occured.
        
        jmp .end
.wrong:
        mov si, [str_idx.wrong_passwd]
        call error_box
.end:
        ret

;=============================================================================
;change_record_passwd ---- change the record password
;=============================================================================
change_record_passwd:
        cmp byte [good_record_number], 0
        je .end

        call  confirm_security_passwd
        jc .end

;        call draw_screen

        mov si, [str_idx.new_record_passwd]
        call input_passwd
        jc .end
        mov bx, ax
        mov cx, dx

        push bx
        push cx
        mov si, [str_idx.retype_passwd]
        call input_passwd
        pop cx
        pop bx
        jc .end
        cmp bx, ax
        jne .wrong
        cmp cx, dx
        jne .wrong

        call get_record_pointer
        mov [si+struc_bootrecord.password], bx
        mov [si+struc_bootrecord.password+2], cx
        
        mov si, [str_idx.passwd_changed]
        call info_box

        inc byte [change_occured]               ; some changes occured.
        
        ret
        
.wrong:
        mov si, [str_idx.wrong_passwd]
        call error_box
.end:
        ret

;=============================================================================
;set_default_record ---- set the default boot record
;=============================================================================
set_default_record:
        cmp byte [good_record_number], 0
        je .end

        call confirm_root_passwd
        jc .end

        mov dl, [focus_record]                  ;
        xor dh, dh                              ; get real record index
        lea si, [good_record_list]              ;
        add si, dx                              ;
        lodsb                                   ;
        
        mov [default_record], al
        inc byte [change_occured]               ; some changes occured.
.end:
        ret

;=============================================================================
;unset_default_record ---- unset the default boot record
;=============================================================================
unset_default_record:
        call confirm_root_passwd
        jc .end

        mov byte [default_record], 0xFF
        inc byte [change_occured]               ; some changes occured.
.end:
        ret
        
;=============================================================================
;toggle_auto_active ---- toggle the auto active switch
;=============================================================================
toggle_auto_active:
        cmp byte [good_record_number], 0
        je .end

        call get_record_pointer
        call check_allow_act
        jc .end

        push si
        call confirm_security_passwd
        pop si
        jc .end
        
        xor byte [si + struc_bootrecord.flags], INFOFLAG_AUTOACTIVE
        inc byte [change_occured]               ; some changes occured.
        
.end:
        ret

;=============================================================================
;toggle_auto_hide ---- toggle the auto hide switch
;=============================================================================
toggle_auto_hide:
        cmp byte [good_record_number], 0
        je .end

        call get_record_pointer
        call check_allow_hide
        jc .end

        push si
        call confirm_security_passwd
        pop si
        jc .end
        
        xor byte [si + struc_bootrecord.flags], INFOFLAG_AUTOHIDE
        inc byte [change_occured]               ; some changes occured.

.end:
        ret

;=============================================================================
;mark_active ---- mark the record active
;=============================================================================
mark_active:
        cmp byte [good_record_number], 0
        je .end

        call get_record_pointer
        call check_allow_act
        jc .end

        push si
        call confirm_security_passwd
        pop si
        jc .end

        mov dl, [si + struc_bootrecord.drive_id]

        push si
        lea si, [good_record_list]
        lea di, [boot_records]

        mov cl, [good_record_number]
        xor ch, ch
        mov dh, SIZE_OF_BOOTRECORD
        xor ebx, ebx
        cld
.loop_clear_act:                                ; clear all active marks of
        push di                                 ; the boot records in same
        lodsb                                   ; drive and father partition.
        mul dh
        add di, ax
        cmp dl, [di + struc_bootrecord.drive_id]
        jne .do_nothing
        cmp [di + struc_bootrecord.father_abs_addr], ebx
        jne .do_nothing
        and byte [di + struc_bootrecord.flags], ~ INFOFLAG_ACTIVE
.do_nothing:
        pop di
        loop .loop_clear_act

        pop si
        
        call mark_record_active
        jc .error                                ; mark active ok
        call recheck_same_records                ; recheck same records
        jc .disk_error
        jmp .end

.error:
        or ax, ax
        jz .cannot_act

.disk_error:
        mov si, [str_idx.disk_error]
        jmp .report_error

.cannot_act:
        mov si, [str_idx.mark_act_failed]
.report_error:
        call error_box
.end:
        ret

;=============================================================================
;toggle_hidden ---- toggle a record's hidden attribute
;=============================================================================
toggle_hidden:
        cmp byte [good_record_number], 0
        je .end

        call get_record_pointer
        call check_allow_hide
        jc .end

        push si
        call confirm_security_passwd
        pop si
        jc .end

        call toggle_record_hidden
        jc .error                                 ; toggle hidden ok
        call recheck_same_records                 ; recheck same records
        jc .disk_error
        jmp .end

.error:
        or ax, ax
        jz .cannot_hide

.disk_error:
        mov si, [str_idx.disk_error]
        jmp .report_error

.cannot_hide:
        mov si, [str_idx.toggle_hid_failed]
.report_error:
        call error_box
.end:
        ret

;=============================================================================
;delete_record ---- delete a boot record
;=============================================================================
delete_record:
        cmp byte [good_record_number], 0
        je .end

        call confirm_root_passwd
        jc .end_clean

        call get_record_pointer
        test byte [si + struc_bootrecord.flags], INFOFLAG_HIDDEN
        jz .del_it

        call toggle_record_hidden           ; unhide it first.
        jnc .del_it                         ; unhide ok, del it.

        or ax, ax
        jz .cannot_hide
        mov si, [str_idx.disk_error]
        jmp .report_error

.del_it:
        xor al, al
        mov di, si
        mov cx, SIZE_OF_BOOTRECORD
        cld
        rep stosb

        inc byte [change_occured]               ; some changes occured.

        call init_good_record_list
        mov al, [good_record_number]
        or al, al
        jz .no_record
        
        dec al
        cmp al, [first_visible_rec]                     ; adjust the cursor
        jae .check_focus_pos                            ; and menu position.
        mov [first_visible_rec], al                     ;
.check_focus_pos:                                       ;
        cmp al, [focus_record]                          ;
        jae .end_clean                                  ;
        mov [focus_record], al                          ;
        jmp .end_clean                                  ;
        
.no_record:
        mov [first_visible_rec], al
        mov [focus_record], al
        jmp .end_clean
        
.cannot_hide:
        mov si, [str_idx.toggle_hid_failed]
.report_error:
        call error_box
        ret
.end_clean:
;        call draw_screen
.end:
        ret
        

;=============================================================================
;rescan_all_drives ---- research all drives for boot records
;=============================================================================
rescan_all_drives:
        or byte [kernel_flags], KNLFLAG_FLOPPY_OK
        jmp rescan_records
        
;=============================================================================
;rescan_fixed_drives ---- research fixed drives for boot records
;=============================================================================
rescan_fixed_drives:
        and byte [kernel_flags], ~ KNLFLAG_FLOPPY_OK
        
;=============================================================================
;rescan_records ---- research all drives for boot records
;=============================================================================
rescan_records:
        call confirm_root_passwd
        jc .end

        mov cl, [good_record_number]
        xor ch, ch
        or cl, cl                               ; if no good record then go to
        jz .init_it                             ; init directly.
        
        lea di, [good_record_list]
        mov dl, SIZE_OF_BOOTRECORD

; unhide all hidden partition first.
.loop_unhide:
        mov al, [di]
        inc di
        mul dl
        lea si, [boot_records]
        add si, ax
        test byte [si + struc_bootrecord.flags], INFOFLAG_HIDDEN
        jz .not_hidden
        call toggle_record_hidden
        jc .hidden_error
        call recheck_same_records
        jc .disk_error

.not_hidden:
        loop .loop_unhide
        
.init_it:
        inc byte [change_occured]               ; some changes occured.

        call init_boot_records
        call init_good_record_list

        xor al, al
        mov byte [focus_record], al
        mov byte [first_visible_rec], al
;        call draw_screen
        jmp .end
        
.hidden_error:
        or ax, ax
        jz .cannot_hide
.disk_error:
        mov si, [str_idx.disk_error]
        jmp .report_error
.cannot_hide:
        mov si, [str_idx.toggle_hid_failed]
.report_error:
        call error_box
.end:
        ret

        
;=============================================================================
;hide_auto_hides ---- hide all partitions that marked auto hide,
;                     except the focus record.
;input:
;      none
;output:
;      cf = 0 success
;      cf = 1 failed
;=============================================================================
hide_auto_hides:
        mov cl, [good_record_number]
        xor ch, ch
        or cl, cl                               ; if no good record then go to
        jz .end_ok                              ; init directly.
        
        lea di, [good_record_list]
        mov dl, SIZE_OF_BOOTRECORD

; hide all auto hide partitions.
.loop_hide:
        mov al, [di]
        inc di

        cmp ch, [focus_record]                  ; do not hide the focus record.
        je .not_hide

        mul dl
        lea si, [boot_records]
        add si, ax
        mov al, [si + struc_bootrecord.flags]
        test al, INFOFLAG_AUTOHIDE
        jz .not_hide
        test al, INFOFLAG_HIDDEN
        jnz .not_hide
        call toggle_record_hidden
        jc .hidden_error
        call recheck_same_records
        jc .disk_error

.not_hide:
        inc ch
        cmp ch, cl
        jb .loop_hide
        
.end_ok:
        clc
        ret
        
.hidden_error:
        or ax, ax
        jz .cannot_hide
.disk_error:
        mov si, [str_idx.disk_error]
        jmp .report_error
.cannot_hide:
        mov si, [str_idx.toggle_hid_failed]
.report_error:
        call error_box
.end:
        stc
        ret

;=============================================================================
;move_main_win_left ---- move the main window left
;=============================================================================
move_main_win_left:
        mov al, [position.main_win_col]
        or al, al
        jz .no_move
        dec al
        mov [position.main_win_col], al
        call draw_screen
.no_move:
        ret

;=============================================================================
;move_main_win_up ---- move the main window up
;=============================================================================
move_main_win_up:
        mov al, [position.main_win_row]
        cmp al, [size.copyright]
        jbe .no_move
        dec al
        mov [position.main_win_row], al
        call draw_screen
.no_move:
        ret

;=============================================================================
;move_main_win_right ---- move the main window right
;=============================================================================
move_main_win_right:
        mov al, [position.main_win_col]
        mov ah, [screen_width]
        sub ah, MAIN_WIN_WIDTH+2
        cmp al, ah
        jae .no_move
        inc al
        mov [position.main_win_col], al
        call draw_screen
.no_move:
        ret

;=============================================================================
;move_main_win_down ---- move the main window down
;=============================================================================
move_main_win_down:
        mov al, [position.main_win_row]
        mov ah, [screen_height]
;        sub ah, [size.main_win_height]
        sub ah, [size.hint]
        cmp al, ah
        jae .no_move
        inc al
        mov [position.main_win_row], al
        call draw_screen
.no_move:
        ret

;=============================================================================
;set_delay_time ---- set the delay time
;=============================================================================
set_delay_time:
        call confirm_root_passwd
        jc .end

;        call draw_screen

        xor ah, ah
        mov al, [color.input_box_msg]
        mov bx, [color.input_box_frame]
        mov cl, 3
        mov si, [str_idx.delay_time]
        lea di, [tmp_buffer]
        mov byte [di], 0
        call input_box
        jc .end_clean

        mov si, di
        mov cl, ch
        xor ch, ch
        call atoi

        cmp ax, 255
        jbe .set_time
        mov al, 255
.set_time:
        mov [delay_time], al
        inc byte [change_occured]               ; some changes occured.
        
.end_clean:
        call draw_screen
.end:
        ret

;=============================================================================
;boot_default ---- boot the default record
;=============================================================================
boot_default:
        mov ah, [default_record]
        lea si, [good_record_list]
        mov cl, [good_record_number]
        xor ch, ch
        or cl, cl
        jz .no_default
        cld
.loop_search:
        lodsb
        cmp al, ah
        je .found_it
        loop .loop_search
        
.no_default:                                ; no default record, do nothing.
        ret
        
.found_it:
        call do_boot_record
        ret
        
;=============================================================================
;boot_it ---- boot the selected record
;=============================================================================
boot_it:
        cmp byte [good_record_number], 0
        je .end

        call confirm_record_passwd
        jc .end

        lea si, [good_record_list]
        mov al, [focus_record]
        xor ah, ah
        add si, ax
        lodsb

        call do_boot_record
.end:
        ret

;=============================================================================
;do_boot_record ---- really boot the given record.
;input:
;      al =  the boot record number.
;=============================================================================
do_boot_record:
        mov bl, SIZE_OF_BOOTRECORD
        mul bl

        lea si, [boot_records]
        add si, ax

        push si
        call ask_save_changes                   ; ask if save the changes.
        call hide_auto_hides
        pop si
        jc .end
        
        call boot_the_record
        
        or al, al
        jnz .disk_error

        mov si, [str_idx.no_system]
        jmp .report_error
.disk_error:
        mov si, [str_idx.disk_error]
.report_error:
        call error_box
.end:
        ret
        
;=============================================================================
;reture_to_bios ---- give control back to BIOS
;=============================================================================
return_to_bios:
        call confirm_root_passwd
        jc .end

        call ask_save_changes                   ; ask if save the changes.
        
        call reset_video_mode
        
%ifdef EMULATE_PROG
        mov ax, 0x4c00                          ; exit to dos
        int 0x21                                ;
%else
        int 0x18                                ; return to BIOS
%endif

.end:
        ret


;=============================================================================
; Duplicate the boot record
;=============================================================================
dup_bootrecord:
        call confirm_root_passwd
        jc .end

        mov ax, SIZE_OF_BOOTRECORD
        mov cx, MAX_RECORD_NUM
        lea di, [boot_records]

.search_empty_slot:
        test byte [di + struc_bootrecord.flags], INFOFLAG_DRIVEOK
        jz .found_empty
        add di, ax
        loop .search_empty_slot
        jmp .no_empty_slot

.found_empty:

        inc byte [change_occured]

        call get_record_pointer
        mov cx, ax 
        cld
        rep movsb
        call init_good_record_list

.no_empty_slot:

.end:
        ret



;=============================================================================
; move the boot record down 
;=============================================================================
move_record_down:
;        call confirm_root_passwd
;        jc .end

        mov al, [focus_record]
        inc al
        cmp al, [good_record_number]
        jae .end
        mov [focus_record], al
        call swap_records
        dec byte [focus_record]
        call down_cursor
.end:
        ret

;=============================================================================
; move the boot record up
;=============================================================================
move_record_up:
;        call confirm_root_passwd
;        jc .end

        mov al, [focus_record]
        or al, al
        jz .end
        call swap_records
        call up_cursor
.end:
        ret

;=============================================================================
; swap current and previous boot record
;=============================================================================
swap_records:
        dec byte [focus_record]
        call get_record_pointer
        mov cx, SIZE_OF_BOOTRECORD
        lea di, [tmp_records_buf]

        cld

        push cx
        push si
        push di
        rep movsb                         ; copy prev bootrecord to tmp
        pop si                            ; si -> tmp
        pop di                            ; di -> prev bootrecord
        pop cx

        inc byte [focus_record]
        call get_record_pointer           ; si -> cur bootrecord
        push cx
        push si
        rep movsb                         ; copy cur bootrecord to prev
        pop di                            ; di -> cur bootrecord
        pop cx

        lea si, [tmp_records_buf]
        rep movsb                         ; copy tmp -> cur bootrecord
;        call init_good_record_list
        inc byte [change_occured]
.end:
        ret



;=============================================================================
;toggle_schedule ---- toggle the schedule of the bootrecord
;=============================================================================
toggle_schedule:
        cmp byte [good_record_number], 0
        je .end

        call confirm_security_passwd
        jc .end

        call get_record_pointer
        test byte [si + struc_bootrecord.ext_flags], EXTFLAG_SCHEDULED
        jnz .clear_schedule

        push si
        call input_schedule_time
        pop si
        jc .end

        mov [si + struc_bootrecord.begin_time], ax
        mov [si + struc_bootrecord.end_time], bx
        or byte [si + struc_bootrecord.ext_flags], EXTFLAG_SCHEDULED
        jmp .end_ok

.clear_schedule:
        and byte [si + struc_bootrecord.ext_flags], ~EXTFLAG_SCHEDULED

.end_ok:
        inc byte [change_occured]
.end:
        ret

;=============================================================================
;input_schedule_time ---- input the schedule time
;input:
;      none
;output:
;      cf = 0 success, 
;           ax = begin time
;           bx = end time
;      cf = 1 cancel
;=============================================================================
input_schedule_time:
        xor ah, ah
        mov al, [color.input_box_msg]
        mov bx, [color.input_box_frame]
        mov si, [str_idx.input_schedule]
        mov cl, 11
        lea di, [tmp_schedule_input]
        mov [di], ah
        
        call input_box
        jnc .input_ok
        jmp .end

.input_ok:
        push cx
        push di
        call draw_screen
        pop di
        pop cx

        cmp ch, cl                          ; the string must be 11 chars long
        jne .invalid_input                  ; (plus 1 zero ending char)
        cmp byte [di + 2], ':'              ; the string must be hh:mm-hh:mm
        jne .invalid_input                  ;
        cmp byte [di + 5], '-'              ;
        jne .invalid_input                  ;
        cmp byte [di + 8], ':'              ; 
        jne .invalid_input                  ;

;convert begin time
        xor ch, ch

        mov si, di
        mov cl, 2
        call atoi
        cmp al, 24                          ; hh must be less than 24
        ja .invalid_input

        mov bl, al
        add si, 3
        call atoi
        cmp al, 60                          ; mm must be less than 60
        jae .invalid_input
        mov cl, al

        mov al, 60
        mul bl
        add ax, cx
        cmp ax, 24*60                       ; begin time must be no more than
        ja .invalid_input                   ; 24*60 minutes

        mov [tmp_schedule_begin], ax

;convert end time
        add si, 3
        mov cl, 2
        call atoi
        cmp al, 24                          ; hh <= 24
        ja .invalid_input

        mov bl, al
        add si, 3
        call atoi
        cmp al, 60                          ; mm < 60
        jae .invalid_input
        mov cl, al

        mov al, 60
        mul bl
        add ax, cx
        cmp ax, 24*60
        ja .invalid_input
        cmp ax, [tmp_schedule_begin]        ; end time must greater than start
        jbe .invalid_input                  ; time

        mov bx, ax
        mov ax, [tmp_schedule_begin]
        clc
        ret

.invalid_input:
        mov si, [str_idx.invalid_schedule]
        call error_box
        stc
.end:
        ret


;=============================================================================
;do_schedule ---- implement the schedule table
;input:
;      none
;output:
;      default_record set to the scheduled record
;=============================================================================
do_schedule:
        pusha
        call get_realtime
        jc .end

        xor cx, cx
        lea si, [boot_records]

.loop_check:
        test byte [si + struc_bootrecord.ext_flags], EXTFLAG_SCHEDULED
        jz .check_next

        call check_bootrecord
        jc .check_next

        cmp ax, [si + struc_bootrecord.begin_time]
        jb .check_next
        cmp ax, [si + struc_bootrecord.end_time]
        ja .check_next

        mov [default_record], cl

.check_next:
        inc cl
        add si, SIZE_OF_BOOTRECORD
        cmp cl, MAX_RECORD_NUM
        jb .loop_check

.end:
        popa
        ret

;=============================================================================
;toggle_keystrokes ---- toggle the keystrokes switch of the bootrecord
;=============================================================================
toggle_keystrokes:
        cmp byte [good_record_number], 0
        je .end

        call confirm_security_passwd
        jc .end

        call get_record_pointer
        test byte [si + struc_bootrecord.ext_flags], EXTFLAG_HAVEKEYS
        jnz .clear_keys

        lea di, [si + struc_bootrecord.keystrokes]
        mov cl, MAX_KEYSTROKES
        push si
        call input_keystrokes
        pop si
        or ch, ch
        jz .end

        or byte [si + struc_bootrecord.ext_flags], EXTFLAG_HAVEKEYS
        jmp .end_ok

.clear_keys:
        and byte [si + struc_bootrecord.ext_flags], ~EXTFLAG_HAVEKEYS

.end_ok:
        inc byte [change_occured]
.end:
        ret


;=============================================================================
;input_keystrokes ---- input a set of key strokes
;input:
;      cl = max key strokes number
;      es:di -> the buffer
;output:
;      es:di -> the buffer filled by key strokes
;      ch = number of key strokes that inputed
;=============================================================================
input_keystrokes:

        call turnon_scrolllock

        push di

        xor ch, ch
        xor ax, ax

        cld

.loop_input:
        push cx
        push di
        call draw_ikm_box
        call read_keystroke
        pop di
        pop cx
        stosw
        jc .end

        inc ch
        cmp ch, cl
        jb .loop_input

.end:
        push cx
        call draw_screen
        pop cx
        pop di

        call turnoff_scrolllock

        ret

;=============================================================================
;draw_ikm_box ---- draw input keystrokes message box
;input:
;       ax = key code
;       ch = key count
;=============================================================================
draw_ikm_box:
        pusha

        mov si, [str_idx.input_keystrokes]
        lea di, [tmp_buffer]

        push di
        push cx

        call strcpy

        mov cl, 4
        call htoa                          ; fill in the key code string
        add di, 4

        mov si, [str_idx.key_count]
        call strcpy

        pop cx
        mov al, ch
        xor ah, ah
        mov cl, 2
        call itoa                          ; fill in the key cound string

        pop si

        call count_lines
        mov dl, cl
        add dl, [size.box_width]
        mov dh, [size.box_height]
        add dh, ch
        mov al, [color.input_box_msg]
        mov bx, [color.input_box_frame]
        mov di, [str_idx.input]
        xchg di, si
        call message_box

        popa
        ret


;=============================================================================
;recheck_same_records ---- recheck all records that same as given record
;input:
;      ds:si -> record
;output:
;      cf = 0  success
;      cf = 1  failed
;=============================================================================
recheck_same_records:
        pusha
        mov ax, [si + struc_bootrecord.drive_id]
        mov ebx, [si + struc_bootrecord.father_abs_addr]
        mov edx, [si + struc_bootrecord.abs_addr]

        lea si, [boot_records]
        mov cx, MAX_RECORD_NUM

.loop_check:
        test byte [si + struc_bootrecord.flags], INFOFLAG_DRIVEOK
        jz .check_next
        cmp [si + struc_bootrecord.drive_id], ax
        jne .check_next
        cmp [si + struc_bootrecord.father_abs_addr], ebx
        jne .check_next
        cmp [si + struc_bootrecord.abs_addr], edx
        jne .check_next

        call check_bootrecord
        jc .end

.check_next:
        add si, SIZE_OF_BOOTRECORD
        loop .loop_check
        clc
.end:
        popa
        ret


;=============================================================================
;show_record_info ---- show the information of the boot record
;=============================================================================
show_record_info:
        call get_record_pointer
        lea di, [tmp_buffer]

        push word [si + struc_bootrecord.end_time]
        push word [si + struc_bootrecord.begin_time]
        push dword [si + struc_bootrecord.password]
        push word [si + struc_bootrecord.ext_flags]
        mov bx, [si + struc_bootrecord.flags]
        mov ax, [si + struc_bootrecord.drive_id]
        add si, struc_bootrecord.name
        push si
        push ax

;write drive id
        mov si, [str_idx.drive_id]
        call strcpy

        xor ah, ah
        mov cx, 3
        call itoa
        add di, cx

;write part id
        mov si, [str_idx.part_id]
        call strcpy

        pop ax
        mov al, ah
        xor ah, ah
        mov cx, 2
        call itoa
        add di, cx

;write part type
        mov si, [str_idx.part_type]
        call strcpy

        mov al, bh
        call get_parttype_str
        call strcpy

;write part_name
        mov si, [str_idx.part_name]
        call strcpy
        pop si
        call strcpy

        mov ax, [str_idx.yes]
        mov dx, [str_idx.no]

;write auto active flag
        mov si, [str_idx.auto_active]
        call strcpy
        test bl, INFOFLAG_AUTOACTIVE
        jz .no_autoact
        mov si, ax
        jmp .autoact
.no_autoact:
        mov si, dx
.autoact:
        call strcpy

;write active flag
        mov si, [str_idx.active]
        call strcpy
        test bl, INFOFLAG_ACTIVE
        jz .no_act
        mov si, ax
        jmp .act
.no_act:
        mov si, dx
.act:
        call strcpy

;write auto hide flag
        mov si, [str_idx.auto_hide]
        call strcpy
        test bl, INFOFLAG_AUTOHIDE
        jz .no_autohide
        mov si, ax
        jmp .autohide
.no_autohide:
        mov si, dx
.autohide:
        call strcpy

;write hidden flag
        mov si, [str_idx.hidden]
        call strcpy
        test bl, INFOFLAG_HIDDEN
        jz .no_hidden
        mov si, ax
        jmp .hidden
.no_hidden:
        mov si, dx
.hidden:
        call strcpy

;write removable flag
        mov si, [str_idx.removable]
        call strcpy
        test bl, INFOFLAG_REMOVABLE
        jz .no_mov
        mov si, ax
        jmp .mov
.no_mov:
        mov si, dx
.mov:
        call strcpy

;write logical flag
        mov si, [str_idx.logical]
        call strcpy
        test bl, INFOFLAG_LOGICAL
        jz .no_log
        mov si, ax
        jmp .log
.no_log:
        mov si, dx
.log:
        call strcpy

;write key strokes flag
        mov si, [str_idx.key_strokes]
        call strcpy
        pop bx
        test bl, EXTFLAG_HAVEKEYS
        jz .no_keys
        mov si, ax
        jmp .keys
.no_keys:
        mov si, dx
.keys:
        call strcpy

;write password flag
        mov si, [str_idx.password]
        call strcpy
        pop ecx
        or ecx, ecx
        jz .no_pswd
        mov si, ax
        jmp .pswd
.no_pswd:
        mov si, dx
.pswd:
        call strcpy


;write schedule time
        mov si, [str_idx.schedule]
        call strcpy
        mov cl, bl
        pop ax
        pop bx

        test cl, EXTFLAG_SCHEDULED
        jz .no_sched
        call schedule_to_str
        jmp .show_info

.no_sched:
        mov si, dx
        call strcpy

.show_info:
        lea si, [tmp_buffer]
        call info_box
        ret
       
;=============================================================================
; do_power_off ---- turn of the power
;=============================================================================
do_power_off:
        mov ax, 3
        call power_off
        ret

;=============================================================================
; utilitiy functions
;=============================================================================

;=============================================================================
;schedule_to_str ---- convert schedule time to string
;input:
;       ax = start time
;       bx = stop time
;       es:di -> buffer
;output:
;       none
;=============================================================================
schedule_to_str:
        pusha
        cld
        call time_to_str
        add di, 5
        mov al, '-'
        stosb
        mov ax, bx
        call time_to_str
        popa
        ret


;=============================================================================
;time_to_str ---- convert time in minute info string hh:mm
;input:
;       ax = time
;       es:di -> buffer
;output:
;       none
;=============================================================================
time_to_str:
        pusha

        mov dl, 60
        div dl
        mov cx, 2
        push ax

        xor ah, ah
        call itoa

        mov al,':'
        add di, cx
        stosb

        pop ax
        mov al, ah
        xor ah, ah
        call itoa
        popa
        ret


;=============================================================================
;read_keystroke ---- read a key stroke
;input:
;      none (the scroll lock must be turned on before read keystrokes)
;output:
;      cf = 0 success, ax = key code
;      cf = 1 cancel (scroll lock was turned off)
;=============================================================================
read_keystroke:
.loop_read:
        mov ah, 0x01
        or ah, [keyboard_type]
        int 0x16
        jnz .have_key
        mov ah, 0x02
        or ah, [keyboard_type]
        int 0x16
        test al, kbScrollMask
        jz .cancel
        jmp .loop_read

.have_key:
        mov ah, [keyboard_type]
        int 0x16
        clc
        ret

.cancel:
        xor ax, ax
        stc
        ret

;=============================================================================
;get_realtime ---- get the machine real time in minutes
;input:
;      none
;output:
;      cf = 0 success, ax = real time in minutes
;      cf = 1 failed
;=============================================================================
get_realtime:
        mov ah, 0x02
        int 0x1a
        jc .end

        xor ah, ah
;cnvert hour from BCD to hex
        mov dh, ch
        shr dh, 4
        mov al, 10
        mul dh
        and ch, 0x0f
        add al, ch
        mov ch, al

;convert minute from BCD to hex
        mov dh, cl
        shr dh, 4
        mov al, 10
        mul dh
        and cl, 0x0f
        add al, cl
        mov cl, al
        
;convert hour and minute into minute
        mov al, 60
        mul ch
        xor ch, ch
        add ax, cx

        clc
.end:
        ret

;=============================================================================
;get_record_pointer ---- get current record pointer
;input:
;      none
;output:
;      ds:si -> record pointer
;=============================================================================
get_record_pointer:
        push dx
        push ax
        mov dl, [focus_record]                  ;
        xor dh, dh                              ; get real record index
        lea si, [good_record_list]              ;
        add si, dx                              ;
        lodsb                                   ;

        mov dl, SIZE_OF_BOOTRECORD              ; get the pointer to
        mul dl                                  ; the record.
        lea si, [boot_records]                  ;
        add si, ax                              ;
        pop ax
        pop dx
        ret
        
;=============================================================================
;confirm_root_passwd ---- confirm the root password
;input:
;      none
;output:
;      cf = 0 success
;      cf = 1 failed or cancel
;=============================================================================
confirm_root_passwd:
        mov bx, [root_password]
        mov cx, [root_password+2]
        or bx, bx
        jnz .have_password
        or cx, cx
        jnz .have_password
        jmp .auth_ok
        
.have_password:                                     
        mov si, [str_idx.root_passwd]               ; check root
        call confirm_passwd                         ; password
.auth_ok:
        ret


;=============================================================================
;confirm_security_passwd ---- confirm the secure password
;                           in normal mode it confirms record password
;                           in security mode it confirms root password
;=============================================================================
confirm_security_passwd:
        test byte [kernel_flags], KNLFLAG_SECURITY
        jz confirm_record_passwd
        jmp confirm_root_passwd

;=============================================================================
;confirm_security_passwd ---- confirm the record password
;=============================================================================
confirm_record_passwd:
        call get_record_pointer
        mov bx, [si + struc_bootrecord.password]
        mov cx, [si + struc_bootrecord.password+2]
        or bx, bx
        jnz .have_password
        or cx, cx
        jnz .have_password
        jmp .auth_ok
        
.have_password:
        mov si, [str_idx.record_passwd]             ; check record
        call confirm_passwd                         ; password
.auth_ok:
        ret

;=============================================================================
;confirm_passwd ---- let user input a password and confirm it.
;input:
;      bx:cx = password
;      ds:si -> message string
;output:
;      cf = 0 success
;      cf = 1 failed or cancel
;=============================================================================
confirm_passwd:
        mov al, [root_login]                        ; check if root
        or al, al                                   ; has login
        jnz .ok                                ;

        call input_passwd
        jc .cancel

        cmp bx, ax
        jne .cmp_root
        cmp cx, dx
        jne .cmp_root
        jmp .ok
        
.cmp_root:
        cmp [root_password], ax
        jne .failed
        cmp [root_password+2], dx
        jne .failed
.ok:
        clc
        ret
        
.failed:
        mov si, [str_idx.wrong_passwd]
        call error_box
        
.cancel:
        stc
        ret


;=============================================================================
;input_passwd ---- input a passwd
;input:
;      ds:si -> message string
;output:
;      cf = 0 success, ax:dx = password
;      cf = 1 cancel
;=============================================================================
input_passwd:
        push bx
        push cx
        
        mov ah, 1
        mov al, [color.input_box_msg]
        mov bx, [color.input_box_frame]
        mov cl, MAX_PASSWORD_LENGTH
        lea di, [tmp_buffer]
        mov byte [di], 0
        call input_box
        jc .cancel_input

        mov si, di
        mov cl, ch
        xor ch, ch
        call calc_password
        clc
.cancel_input:
        push ax
        push dx
        pushf
        call draw_screen
        popf
        pop dx
        pop ax
        pop cx
        pop bx
        ret

;=============================================================================
;error_box ---- draw error message box.
;input:
;      ds:si -> error message
;output:
;      none
;=============================================================================
error_box:
        call count_lines
        mov dl, cl
        add dl, [size.box_width]
        mov dh, [size.box_height]
        add dh, ch

        mov al, [color.error_box_msg]
        mov bx, [color.error_box_frame]
        mov di, [str_idx.error]
        xchg di, si
        call message_box
        xor ah, ah
        int 0x16
        push ax
        call draw_screen
        pop ax
        ret

;=============================================================================
;info_box ---- draw infomation message box.
;input:
;      ds:si -> infomation message
;output:
;      none
;=============================================================================
info_box:
        call count_lines
 
        mov dl, cl
        add dl, [size.box_width]
        mov dh, [size.box_height]
        add dh, ch

        mov al, [color.info_box_msg]
        mov bx, [color.info_box_frame]
        mov di, [str_idx.info]
        xchg di, si
        call message_box
        xor ah, ah
        int 0x16
        push ax
        call draw_screen
        pop ax
        ret

;=============================================================================
;include area
;=============================================================================

%include "ui.asm"
%include "utils.asm"
%include "knl.asm"
%include "hd_io.asm"

;=============================================================================
; data area
;=============================================================================

func_list:
.hotkeys:
        dw  kbUp
        dw  kbEnhUp
        dw  kbDown
        dw  kbEnhDown
        dw  kbEnter
        dw  kbEnhEnter
        dw  kbEsc
        dw  kbF1
        dw  kbCtrlF1
        dw  kbF2
        dw  kbF3
        dw  kbF4
        dw  kbF5
        dw  kbF6
        dw  kbF7
        dw  kbF8
        dw  kbShiftF8
        dw  kbF9
        dw  kbF10
        dw  kbCtrlF10
        dw  kbAltF10
        dw  kbGrayPlus
        dw  kbDel
        dw  kbEnhDel
        dw  kbHome
        dw  kbEnhHome
        dw  kbPgDn
        dw  kbEnhPgDn
        dw  kbEnd
        dw  kbEnhEnd
        dw  kbCtrlD
        dw  kbCtrlI
        dw  kbCtrlH
        dw  kbCtrlT
        dw  kbCtrlQ
        dw  kbCtrlP
        dw  kbCtrlU
        dw  kbCtrlN
        dw  kbCtrlS
        dw  kbCtrlK
        dw  kbSlash
        dw  kbQuestion
        dw  kbCtrlF12

.entries:
        dw  up_cursor                       ; kbUp
        dw  up_cursor                       ; kbUp
        dw  down_cursor                     ; dbDown
        dw  down_cursor                     ; dbDown
        dw  boot_it                         ; kbEnter
        dw  boot_it                         ; kbEnter
        dw  boot_default                    ; kbEsc
        dw  show_help                       ; kbF1
        dw  show_about                      ; kbCtrlF1
        dw  save_changes                    ; kbF2
        dw  change_name                     ; kbF3
        dw  mark_active                     ; kbF4
        dw  toggle_hidden                   ; kbF5
        dw  toggle_auto_active              ; kbF6
        dw  toggle_auto_hide                ; kbF7
        dw  set_default_record              ; kbF8
        dw  unset_default_record            ; kbShiftF8
        dw  change_record_passwd            ; kbF9
        dw  change_root_passwd              ; kbF10
        dw  login_as_root                   ; kbCtrlF10
        dw  change_security_mode            ; kbAltF10
        dw  change_video_mode               ; kbGrayPlus
        dw  move_main_win_left              ; kbDel
        dw  move_main_win_left              ; kbDel
        dw  move_main_win_up                ; kbHome
        dw  move_main_win_up                ; kbHome
        dw  move_main_win_right             ; kbPgDn
        dw  move_main_win_right             ; kbPgDn
        dw  move_main_win_down              ; kbEnd
        dw  move_main_win_down              ; kbEnd
        dw  delete_record                   ; kbCtrlD
        dw  rescan_all_drives               ; kbCtrlI
        dw  rescan_fixed_drives             ; kbCtrlH
        dw  set_delay_time                  ; kbCtrlT
        dw  return_to_bios                  ; kbCtrlQ
        dw  dup_bootrecord                  ; kbCtrlP
        dw  move_record_up                  ; kbCtrlU
        dw  move_record_down                ; kbCtrlN
        dw  toggle_schedule                 ; kbCtrlS
        dw  toggle_keystrokes               ; kbCtrlK
        dw  show_record_info                ; kbSlash
        dw  show_record_info                ; kbQuestion
        dw  do_power_off                    ; kbCtrlF12

;END OF KERNEL
        dw BR_GOOD_FLAG
end_of_kernel:

;=============================================================================
;theme data
;=============================================================================
theme_start:

%ifdef THEME_ZH
%include "themes/theme-zh.asm"
%else
%include "themes/theme-us.asm"
%endif

;=============================================================================
; temp data area
;=============================================================================

%ifndef EMULATE_PROG
times MAX_SBM_SIZE - ($-$$) db 0                ; skip enough space for theme.
%endif

%ifdef EMULATE_PROG
db 'TEMP DATA',0
%endif

%include "tempdata.asm"
db 'END DATA'

