; ui.asm
;
; Functions for User Interface
;
; Copyright (C) 2000, Suzhe. See file COPYING for details.
;


%ifndef MAIN
%include "ui.h"
%include "utils.asm"
%endif

%define HAVE_UI

%define SCR_BUFF_SEG    0xB800
%define BIOS_DATA_SEG   0x0040

%define BIOS_KEYSTAT_OFF 0x0017

      bits 16

%if 1
;=============================================================================
;draw_string_hl ---- Draw a zero ending string with highlighted characters 
;                    at special position
;input:
;      bl = attribute for normal characters
;           high 4 bit Background color and low 4 bit Foreground color
;      bh = attribute for hightlight characters
;      dh = start row
;      dl = start column
;      ds:si -> the string to be displayed
;output:
;      none
;=============================================================================
draw_string_hl:
        pusha
        push dx
        mov cx,1
        cld
.start:

        lodsb
        or al,al
        jz .end

        cmp al,0x0d                ; if need Change row
        jne .no_cr
        pop dx
        inc dh
        push dx
        jmp short .start

.no_cr:
        cmp al, '~'
        jne .draw_it
        xchg bh, bl
        jmp .next_char

.draw_it:
        call draw_char

.next_char:
        inc dl
        jmp short .start
.end:
        pop dx
        popa
        ret
;=============================================================================
%endif

;=============================================================================
;draw_string ---- Draw a zero ending string at special position
;input:
;      bl = high 4 bit Background color and low 4 bit Foreground color
;      dh = start row
;      dl = start column
;      ds:si -> the string to be displayed
;output:
;      none
;=============================================================================
draw_string:
        pusha
        push dx
        mov cx,1
        cld
.start:

        lodsb
        or al,al
        jz .end

        cmp al,0x0d                ; if need Change row
        jne .no_cr
        pop dx
        inc dh
        push dx
        jmp short .start

.no_cr:
        call draw_char

        inc dl
        jmp short .start
.end:
        pop dx
        popa
        ret
;=============================================================================


;=============================================================================
;draw_char ---- Draw chars at special position
;input:
;      bl = high 4 bit Background color and low 4 bit Foreground color
;      dh = start row
;      dl = start column
;      al = the char to be displayed
;      cx = repeat times
;output:
;      none
;=============================================================================
draw_char:
%if 1                                     ; directly write to video buffer
        pusha
        push es
        mov ah, bl
        push ax

        mov ax, SCR_BUFF_SEG
        mov es, ax
        mov al, [screen_width]
        mul dh
        xor dh, dh
        add ax, dx
        shl ax, 1
        mov di, ax

        pop ax
        rep stosw
        pop es
        popa
%else
        mov ah,2
        xor bh,bh
        int 0x10
        mov ah,0x09
        int 0x10
%endif
        ret
;=============================================================================

%if 0
;=============================================================================
;draw_string_tty ---- Draw a string ending by zero ( tty mode )
;input:
;      ds:si -> string
;output:
;      none
;=============================================================================
draw_string_tty:
        pusha
        cld
.draw1:
        lodsb
        or al, al
        jz .end
        mov bx,7
        mov ah,0x0e
        int 0x10
        jmp .draw1
.end:
        popa
        ret
;=============================================================================
%endif

;=============================================================================
;draw_window ---- Draw a framed window
;input:
;      ch = row of top left corner
;      cl = column of top left corner
;      dh = row of bottom right corner
;      dl = column of bottom right corner
;      bl = high 4 bit Background color and low 4 bit Foreground color
;      bh = title attribute (define same as bl)
;      ds:si -> title
;output:
;      none
;=============================================================================
draw_window:
        pusha
        mov [ui_tmp.left_col], cx          ;
        mov [ui_tmp.right_col], dx         ; save window pos and attribute
        mov [ui_tmp.frame_attr], bx        ;

;Clear frame background
        xchg bh,bl
        mov ax,0x0600
        int 0x10

        xchg dx,cx
        mov cx,1

;Draw four corners
        cmp byte [draw_frame_method], 2             ; check draw method.
        jb .draw_top_corner_as_frame
        mov bl, [ui_tmp.title_attr]
        jmp .draw_top_corner
.draw_top_corner_as_frame:
        mov bl, [ui_tmp.frame_attr]
.draw_top_corner:
        mov al, [frame_char.tl_corner]
        call draw_char

        mov dl, [ui_tmp.right_col]
        mov al, [frame_char.tr_corner]
        call draw_char

        mov bl, [ui_tmp.frame_attr]
        mov dh, [ui_tmp.bottom_row]
        mov al, [frame_char.br_corner]
        call draw_char
  
        mov dl, [ui_tmp.left_col]
        mov al, [frame_char.bl_corner]
        call draw_char

;Draw bottom horizontal line
        inc dl
        mov cl, [ui_tmp.right_col]
        sub cl, dl
        mov al, [frame_char.bottom]
        call draw_char

;Draw top horizontal line
        cmp byte [draw_frame_method], 1             ; check draw method.
        jb .draw_top_line_as_frame
        mov bl, [ui_tmp.title_attr]
        jmp .draw_top_line
.draw_top_line_as_frame:
        mov bl, [ui_tmp.frame_attr]
.draw_top_line:
        mov dh, [ui_tmp.top_row]
        mov al, [frame_char.top]
        call draw_char

;Draw title
        call strlen
        or cx,cx
        jz .no_title

        mov al, [ui_tmp.right_col]
        sub al, [ui_tmp.left_col]
        sub al, cl
        inc al
        shr al,1
        mov dl, [ui_tmp.left_col]
        add dl,al
        mov dh, [ui_tmp.top_row]

        mov bl, [ui_tmp.title_attr]
        call draw_string

.no_title:

;Draw vertical line
        mov bl, [ui_tmp.frame_attr]
        mov dh, [ui_tmp.top_row]
        inc dh

        mov cx,1

.draw_vert_line:
        mov al, [frame_char.left]
        mov dl, [ui_tmp.left_col]
        call draw_char
        mov al, [frame_char.right]
        mov dl, [ui_tmp.right_col]
        call draw_char

        inc dh
        cmp dh, [ui_tmp.bottom_row]
        jb .draw_vert_line

;Draw shadow
        mov bl, 0x08
        mov ch, [ui_tmp.bottom_row]
        mov cl, [ui_tmp.left_col]
        inc ch
        inc cl
        inc cl
        mov dh, [ui_tmp.bottom_row]
        mov dl, [ui_tmp.right_col]
        inc dh
        call draw_shadow
        mov ch, [ui_tmp.top_row]
        mov cl, [ui_tmp.right_col]
        inc ch
        inc cl
        inc dl
        inc dl
        call draw_shadow

        popa
        ret

;=============================================================================


;=============================================================================
;draw_shadow_char ---- Draw a shadow char
;input:
;      bl = high 4 bit Background color and low 4 bit Foreground color
;      dh = row
;      dl = column
;      cx = repeat times
;output:
;      none
;=============================================================================
draw_shadow_char:
        xor bh,bh
        mov ah,0x02
        int 0x10
        mov ah,0x08
        int 0x10
        call draw_char
        ret
;=============================================================================

;=============================================================================
;draw_shadow ---- Draw shadow block
;input:
;      ch = row of top left corner
;      cl = column of top left corner
;      dh = row of bottom right corner
;      dl = column of bottom right corner
;      bl = high 4 bit Background color and low 4 bit Foreground color
;output:
;      none
;=============================================================================
draw_shadow:
        pusha
.loop_row:
        push dx
.loop_col:
        push cx
        mov cx,1
        call draw_shadow_char
        pop cx
        dec dl
        cmp cl, dl
        jbe .loop_col
        pop dx
        dec dh
        cmp ch, dh
        jbe .loop_row

        popa
        ret
;=============================================================================

;=============================================================================
;message_box ---- Draw a message box
;input:
;      ds:si -> title
;      ds:di -> message
;      al = message attribute
;      bh = title attribute
;      bl = frame attribute
;      dh = box height
;      dl = box width
;output:
;      none
;=============================================================================
message_box:
        pusha
        push ax
;Calculate Window position
        mov ch, [screen_height]
        sub ch, dh
        shr ch, 1
        mov cl, [screen_width]
        sub cl, dl
        shr cl, 1
        add dh, [screen_height]
        shr dh, 1
        dec dh
        add dl, [screen_width]
        shr dl, 1

;Draw window and message
        push cx
        call draw_window
        pop dx
        add dh, 2
        add dl, 3
        xchg si,di
        pop bx
        or si,si
        jz .end
        call draw_string
.end:
        popa
        ret
;=============================================================================


;=============================================================================
;input_box ---- draw a input box and input a string
;input:
;      ah = input method ( 0 = normal, 1 = security )
;      al = message attribute
;      bh = title attribute
;      bl = frame attribute
;      cl = max input length
;      ds:si -> message ( no more than one line )
;      es:di -> buffer to store input string
;output:
;      cf = 0 , ah = 0 ok, ch = number of inputed character
;      cf = 1 , ah != 0 cancel, ch = 0
;=============================================================================
input_box:
        push ax
        push cx

        mov dh, [size.box_height]               ; dh = box height
        inc dh
        mov dl, cl                              ; dl = box width
        add dl, [size.box_width]

        call strlen                             ; box width = msg length +
        push cx                                 ; max input lenght + 4
        add dl, cl
        
        mov ch, [screen_height]                 ; calculate the coordinate
        sub ch, dh                              ; of input box.
        shr ch, 1                               ; ch = top left row
        mov cl, [screen_width]                  ; cl = top left column
        sub cl, dl                              ; dh = bottom right row
        shr cl, 1                               ; dl = bottom right column
        add dh, [screen_height]                 ;
        shr dh, 1                               ;
        dec dh
        add dl, [screen_width]                  ;
        shr dl, 1                               ;

        push cx                                 ; save top left position
        push si
        mov si, [str_idx.input]                 ; get the title of input box
        call draw_window                        ;
        pop si
        pop dx
        
        add dl, 2                               ;
        add dh, 2                               ; display the input message
        mov bl, al                              ;
        call draw_string                        ;

        pop cx
        add dl, cl
        
        pop cx
        pop ax
        mov bh, ah
        mov bl, 0x0F
        call input_string
        ret
;=============================================================================


;=============================================================================
;input_string ---- input a string from keyboard (zero ending)
;input:
;      bh = input type ( 0=normal, 1=security )
;      bl = text attribute
;      cl = max length
;      dh = start row
;      dl = start column
;      es:di -> buffer
;output:
;      cf = 0, ah = 0 ok, ch = number of inputed character
;      cf = 1, ah != 0 cancel, ch = 0
;=============================================================================
input_string:
        push di
        push si
        cld

; clear the input area
        push bx
        xor ch,ch
        mov al, 0x20
        call draw_char
        
        push cx                                 ; if the buffer is not
        mov si, di                              ; empty then keep the
        call strlen                             ; content and move the
        mov bx, cx                              ; cursor to the end
        pop cx                                  ; of the string.
        xor bh, bh                              ;
        cmp cl, bl                              ;
        ja .longer                              ;
        mov bl, cl                              ;
.longer:                                        ;
        mov byte [si + bx], 0                   ;
        add ch, bl                              ;
        pop bx                                  ;
        push bx                                 ;
        call draw_string                        ;
        
        add dl, ch                              ; adjust the pointer and
        mov al, ch                              ; cursor position
        xor ah, ah                              ;
        add di, ax                              ;
        
        mov ah,2                                ;
        xor bh,bh                               ; set cursor position
        int 0x10                                ;

        call show_cursor

        pop bx

.loop_input:
        xor ah,ah
        int 0x16
        cmp ax, kbEsc
        jz .cancel
        cmp ax, kbEnter
        jz .end_input
        cmp ax, kbBack
        jz .back_space
        cmp ax, kbDel
        jz .back_space
        or al,al
        jz .loop_input

;input a char
        cmp ch,cl
        jae .loop_input

        stosb
        or bh,bh
        jz .normal_input
        mov al, '*'        ;draw a star instead of input char
.normal_input:
        push bx
        push cx
        mov cx,1
        call draw_char
        pop cx
        inc ch
        inc dl

        xor bh, bh                                  ;
        mov ah,2                                    ; adjust the cursor
        int 0x10                                    ;
        
        pop bx
        jmp .loop_input

;delete a char
.back_space:
        or ch,ch
        jz .loop_input

        dec ch
        dec di
        dec dl
        mov al, 0x20
        push cx
        push bx
        mov cx,1
        call draw_char
        
        mov ah, 2                                   ;
        xor bh, bh                                  ; adjust the cursor
        int 0x10                                    ;
        
        pop bx
        pop cx
        jmp .loop_input

.end_input:
        xor ax,ax
        stosb
        call hide_cursor
        jmp .end
.cancel:
        xor ch,ch
        call hide_cursor
        stc
.end:
        pop si
        pop di
        ret
;=============================================================================


;=============================================================================
;set_video_mode ---- Set the Alphabet Video Mode
;input:
;      al = 0 , set screen resolution to 90x25,
;           otherwise set to 80x25
;      bl = character bit size ( 8 or 9 )
;      cx = fonts number
;      es:bp -> fonts data
;output:
;      none
;=============================================================================
BIOS_CRT_COLS        equ 0x4A
BIOS_ADDR_6845       equ 0x63

set_video_mode:
        push es
        push bx
        push ax

        call reset_video_mode

;Establish CRTC vertical timing and cursor position in character matrix
;and set user fonts table
        or cx, cx
        jz .set_res                          ; no font data
        or bp, bp
        jz .set_res                          ; no font data
        
.loop_set_fonts:                             ; set user defined chars
        push cx
        xor cx,cx
        inc cl
        xor dh, dh
        mov dl, [es:bp]
        inc bp
        mov ax,0x1100
        mov bx,0x1000
        int 0x10
        pop cx
        add bp, 16
        loop .loop_set_fonts

.set_res:
        mov byte [screen_width], 80
        mov byte [screen_height], 25

        mov ax,0x40
        mov es,ax
        pop ax

        or al,al
        jnz .skip_res_set

        mov dx,[es:BIOS_ADDR_6845]                   ; CRTC I/O port

;Enable I/O writes to CRTC registers
        mov al,0x11
        out dx,al
        inc dx
        in al,dx
        dec dx
        mov ah,al
        mov al,0x11
        push ax
        and ah,01111111b
        out dx,ax

;Establish CRTC horizontal timing
        lea si, [VideoHorizParams]

        mov cx,7
        
        cld
.set_CRTC:
        lodsw
        out dx,ax
        loop .set_CRTC

;write-protect CRTC registers
        pop ax
        out dx,ax

;Program the Sequencer and Attribute Controller for 9 dots per character
        
        mov dx, 0x3c4
        mov ax, 0x0100
        cli
        out dx,ax

        mov ax,0x0101
        out dx,ax
        mov ax,0x0300
        out dx,ax
        sti

        mov bx,0x0013
        mov ax,0x1000
        int 0x10

        mov byte [screen_width], 90
.skip_res_set:

;Program the Attribute Controller for 8- or 9-bit character codes
        mov ax,0x1000
        mov bx,0x0f12
        pop dx
        cmp dl,8
        je .svm01
        mov bh,7
.svm01:
        int 0x10

;Update video BIOS data area
        mov al,[screen_width]
        mov [es:BIOS_CRT_COLS],al

;Set background highlight attribute
        pop es
        mov ax,0x1003
        xor bl,bl
        int 0x10
        call hide_cursor

        ret
;=============================================================================

;=============================================================================
;hide_cursor ---- Hide the cursor
;input:
;      none
;output:
;      none
;=============================================================================
hide_cursor:
        pusha
        mov ah,1
        mov cx,0x6f00
        int 0x10
        popa
        ret
;=============================================================================

;=============================================================================
;show_cursor ---- Show the cursor
;input:
;      none
;output:
;      none
;=============================================================================
show_cursor:
        pusha
        mov ah,1
        mov cx,0x0e0f
        int 0x10
        popa
        ret
;=============================================================================

;=============================================================================
;reset_video_mode ---- Reset the VideoMode
;input:
;      none
;output:
;      none
;=============================================================================
reset_video_mode
        pusha
        mov ax,3
        int 0x10
        call show_cursor
        popa
        ret
;=============================================================================


;=============================================================================
;draw_icon ---- Draw a icon at special position
;input:
;      dh = start row
;      dl = start column
;      ch = number of row
;      cl = number of column
;      ds:si -> icon data , which is a two dim word array, each elements
;               indicates a char. high byte is the attribute, low byte is
;               is the char code.
;output:
;      none
;=============================================================================
draw_icon:
        or si, si
        jz .end
        or cx, cx
        jz .end
        
        pusha
        cld
.loop_row:
        push dx
        push cx
.loop_col:
        push cx
        mov cx,1
        lodsw
        xchg ah,bl
        call draw_char
        
        pop cx
        inc dl
        dec cl
        jnz .loop_col
        
        pop cx
        pop dx
        inc dh
        dec ch
        jnz .loop_row

        popa
.end:
        ret
;=============================================================================

;=============================================================================
;draw_background ---- Draw the background using specified icon
;input:
;      bh = background color when no icon
;      cx = icon size (ch = row, cl = col)
;      ds:si -> icon data , which is a two dim word array, each elements
;               indicates a char. high byte is the attribute, low byte is
;               is the char code.
;output:
;      none
;=============================================================================
draw_background:
        pusha
        or si,si
        jnz .normal_bg

;no icon. clear background.
        xor cx,cx
        mov dh,[screen_height]
        mov dl,[screen_width]
        dec dh
        dec dl
        mov ax,0x0600
        int 0x10
        popa
        ret

.normal_bg:
        xor dx,dx

.loop_row:
        push dx
.loop_col:
        call draw_icon
        add dl, cl
        cmp dl, [screen_width]
        jb .loop_col
        pop dx
        add dh, ch
        cmp dh, [screen_height]
        jb .loop_row
        popa
        ret
;=============================================================================

;=============================================================================
;turnon_scrolllock ---- turn on the scroll lock key
;input: none
;output: none
;=============================================================================
turnon_scrolllock:
        pusha
        push es
        mov ax, BIOS_DATA_SEG
        mov es, ax
        mov di, BIOS_KEYSTAT_OFF
        or byte [es:di], kbScrollMask
        pop es
        popa
        ret

;=============================================================================
;turnoff_scrolllock ---- turn off the scroll lock key
;input: none
;output: none
;=============================================================================
turnoff_scrolllock:
        pusha
        push es
        mov ax, BIOS_DATA_SEG
        mov es, ax
        mov di, BIOS_KEYSTAT_OFF
        and byte [es:di], ~ kbScrollMask
        pop es
        popa
        ret


%if 0
;=============================================================================
; Command menu functions
;=============================================================================


;=============================================================================
;draw_menu_item ---- draw a menu item on the screen
;input:
;     bl = attribute for normal characters 
;     bh = attribute for highlighted characters
;     cx = item number to be displayed
;     dh = start row
;     dl = start column
;     ds:si - > the string table of menu items
;=============================================================================
draw_menu_item:
        pusha
        add dl, cl
        shl cx
        add si, cx
        mov si, [si]
        call draw_string_hl
        popa
        ret


;=============================================================================
; draw_menu ---- draw a menu on the screen
;=============================================================================
%endif


;=============================================================================
; Data Area
;=============================================================================
VideoHorizParams dw 0x6B00,0x5901,0x5A02,0x8E03,0x5F04,0x8C05,0x2D13 ;8-wide

screen_width      db     90
screen_height     db     25

%ifndef MAIN

; how to draw window frame
draw_frame_method   db  0          ; = 0 means draw all frame using frame attr.
                                   ; = 1 means draw top horizontal line using
                                   ;     title attr.
                                   ; = 2 means draw top corner and horizontal
                                   ;     line using title attr.
frame_char:
.top             db     0x020
.bottom          db     0x0CD
.left            db     0x0BA
.right           db     0x0BA
.tl_corner       db     0x0C9               ; top left corner
.tr_corner       db     0x0BB               ; top right corner
.bl_corner       db     0x0C8               ; bottom left corner
.br_corner       db     0x0BC               ; bottom right corner

size:
.box_height      db  4

str_idx:
.input          dw  string.input

string:
.input        db     'Input',0

%include "tempdata.asm"

%endif

