; utils.asm
;
; Some utility functions
;
; Copyright (C) 2000, Suzhe. See file COPYING for details.
;

%define HAVE_UTILS

;=============================================================================
;itoa ---- convert integer to ascii string (the string is zero ending)
;input:
;      ax = the integer to be converted
;      cl = max length of the integer
;      es:di -> buffer
;output:
;      none
;=============================================================================
itoa:
        pusha
        xor ch, ch
        add di, cx
        mov byte [di], 0
        mov bx, 10
.loop_itoa:
        xor dx, dx
        dec di
        div bx
        add dl, '0'
        mov [di], dl
        dec cx
        or ax, ax
        jz .end_itoa
        or cx, cx
        jnz .loop_itoa
.end_itoa:
        or cx, cx
        jz .end
.loop_fillspace:
        dec di
        mov byte [di], ' '
        loop .loop_fillspace
.end:
        popa
        ret

;=============================================================================
;atoi ---- convert ascii string to integer (the string is zero ending)
;input:
;      cl = length of the integer
;      ds:si -> buffer
;output:
;      ax = convert result ( <= 65535 )
;=============================================================================
atoi:
        pusha
        add si, cx
        xor ax, ax
        xor dx, dx
        xor ch, ch
        mov [tmp_atoi_result], ax
        mov bx, 1
        std
        or cx, cx
        jz .end_atoi
        lodsb
.loop_atoi:
        lodsb
        xor ah, ah
        sub al, '0'
        mul bx
        add [tmp_atoi_result], ax
        
        mov ax, bx                  ;
        shl bx, 3                   ; bx = bx * 10
        shl ax, 1                   ;
        add bx, ax                  ;
        loop .loop_atoi
.end_atoi:
        cld
        popa
        mov ax, [tmp_atoi_result]
        ret

;=============================================================================
;strlen ---- Count Length of a zero ending string
;input:
;      ds:si -> string
;output:
;      cx = length (not include the ending zero)
;=============================================================================
strlen:
        push ax
        xor cx,cx

        or si,si
        jz .end

        push si
        cld
.loop_count:
        inc cx
        lodsb
        or al, al
        jnz .loop_count

        dec cx
        pop si

.end:
        pop ax
        ret

;=============================================================================
;strchr ---- search a char in a string
;input:
;      al = the char to be searched
;      cx = length of the string
;      es:di -> string
;output:
;      cf = 0 the char was found
;            cx = offset of the char
;      cf = 1 the char was not found
;            cx = length of the string
;=============================================================================
strchr:
        push bx
        push di
        mov bx,cx
        cld
        repnz scasb
        pop di
        xchg bx, cx
        jnz .not_found
        sub cx, bx
        dec cx
        pop bx
        clc
        ret
.not_found:
        pop bx
        stc
        ret

;=============================================================================
;strncpy ---- copy strings
;input:
;      cx = max number of chars to be copied
;      ds:si -> source string
;      es:di -> dest string
;output:
;      cx = number of chars actually copied (not include the ending zero char)
;      es:di -> point to next char of the end of dest string and set it to
;               zero
;=============================================================================
strncpy:
        push si
        push ax
        push bx
        mov bx, cx
        cld
.loop_copy:
        lodsb
        stosb
        or al, al
        jz .end
        loop .loop_copy
        xor al, al
        stosb
.end:
        dec di
        xchg bx, cx
        sub cx, bx
        pop bx
        pop ax
        pop si
        ret

;=============================================================================
;strcpy ---- copy strings
;input:
;      ds:si -> source string
;      es:di -> dest string
;output:
;      es:di -> point to next char of the end of dest string and set it to
;               zero
;=============================================================================
strcpy:
        push si
        push ax
        push bx
        cld
.loop_copy:
        lodsb
        stosb
        or al, al
        jz .end
        jmp .loop_copy
        xor al, al
        stosb
.end:
        dec di
        pop bx
        pop ax
        pop si
        ret


;=============================================================================
;calc_password ---- calculate the password
;input:
;      ds:si -> the pasword string (zero ending)
;      cx = max length of the password
;output:
;      ax:dx = the encrypted password (32 bits)
;=============================================================================
calc_password:
        push si
        xor edx, edx
        xor eax, eax
        cld
.loop_calc:
        lodsb
        or al, al
        jz .end
        not al
        rol al, 4
        add edx, eax
        rol edx, 2
        loop .loop_calc
.end:
        mov ax, dx
        ror edx, 16
        pop si
        ret

;=============================================================================
;htoa ---- hex to ascii
;input:
;     ax = hex number
;     cl = length of hex number (1 to 4)
;     es:di -> buffer to store ascii string
;output:
;     es:di -> ascii string
;=============================================================================
htoa:
        pusha
        xor ch, ch
        add di, cx
        mov byte [di], 0

.loop_conv:
        mov bl, al
        and bl, 0x0f
        cmp bl, 0x09
        jbe .be_9
        add bl, 'A'-'0'-10
.be_9:
        add bl, '0'
        dec di
        mov [di], bl
        shr ax, 4
        loop .loop_conv

        popa
        ret

;=============================================================================
;count_lines ---- count how many lines in a string.
;input:
;      ds:si -> string
;output:
;      ch = number of lines
;      cl = max line length
;=============================================================================
count_lines:
        push si
        push bx
        push ax

        cld

        xor cx, cx
        xor bx, bx

        or si, si
        jz .end

        inc ch
.loop_count:
        lodsb
        or al, al
        jz .ending

        cmp al, 0x0d
        je .new_line

        inc bl
        jmp .loop_count

.new_line:
        inc ch

        mov bh, bl
        xor bl, bl
        cmp bh, cl
        jbe .loop_count
        mov cl, bh
        jmp .loop_count

.ending:
        cmp bl, cl
        jbe .end
        mov cl, bl

.end:
        pop ax
        pop bx
        pop si
        ret

;=============================================================================
; power_off ---- turn the power off
;input:
;       none
;output:
;       never return if successful.
;       cf = 1 on error.
;=============================================================================
power_off:
        pusha
        mov ax, 0x5300
        xor bx, bx
        int 0x15                                 ; check if apm present
        jc .end

        mov ax, 0x5301
        xor bx, bx
        int 0x15
        jc .end

        mov ax, 0x5380
        mov bh, 0x8c
        int 0x15

        mov ax, 0x40
        mov bx, 0xd8
        push ds
        mov ds, ax
        or byte [ds:bx], 0x10
        pop ds

        mov ax, 0x5307
        mov bx, 1
        mov cx, 3
        int 0x15

.end:
        popa
        ret


%ifndef MAIN
%include "tempdata.asm"
%endif

